
#include <fstream>

#include <boost/regex.hpp>

#include "jmath/indirectArray.hpp"
#include "jmath/misc.hpp"
#include "jmath/random.hpp"
#include "rtslam/quatTools.hpp"

using namespace jafar;
using namespace jafar::rtslam;
namespace ublasExtra = jafar::jmath::ublasExtra;

#undef JFR_ASSERT

#define JFR_ASSERT(predicate, message) \
{ \
  if (!(predicate)) \
  { \
    std::cerr << message << " (" << #predicate << ")" << std::endl; \
    throw kernel::JafarException(kernel::JafarException::ASSERT,			\
       "",_JFR_MODULE_, __FILE__, __LINE__); \
  } \
}


/*
Donc ici :
Je prends deux fichiers, slam et vérité, et pour chaque timestamp du slam :
- j'interpole la vérité
- je mets à jour la distance parcourue par segments de 10cm
- je calcule l'erreur (moyenne et incertitude)
- optionnellement je recale le facteur d'échelle au t demandé

j'affiche ensuite :
- temps, distance parcourue
- puis d'abord pour vérité, ensuite pour slam : position, orientation euler, norme vitesse, et incertitudes avec cov complete
erreur xy, z, w, v / + uncert

puis je découpe aléatoire la traj en tranches de 20m, je calcule erreur et incertitude,
et la puissance de cette erreur

*/

const size_t ncols = 1 + 3+4+3 + 3+4+3 + 3+6+3;
static const char* col_names[ncols] = { "time", "x", "y", "z", "qw", "qx", "qy", "qz", "vx", "vy", "vz",
"sig_x", "sig_y", "sig_z", "sig_qw", "sig_qx", "sig_qy", "sig_qz", "sig_vx", "sig_vy", "sig_vz",
"cov_xy", "cov_xz", "cov_yz", "cov_qwqx", "cov_qwqy", "cov_qwqz", "cov_qxqy", "cov_qxqz", "cov_qyqz", "cov_vxvy", "cov_vxvz", "cov_vyvz"};

	struct Line
	{
		double t;
		double s;
		jblas::vec ref;
		jblas::vec eval;
	};


bool find_cols(std::ifstream & file, jblas::ind_array & ia, size_t & nc)
{
	std::vector<std::string> header;
	std::string line;
	file.seekg(0, file.beg);
	while(std::getline(file, line))
		if (line.size() >= 2 && line[0] == '#' && line[1] == ' ') header.push_back(line);

	// find total number of cols
	nc = 0; size_t col;
	std::ostringstream ossp1;
	ossp1 << ".*[ \t]([0-9]*):[a-z_]+.*";
	boost::regex pattern1(ossp1.str());
	for(size_t j = 0; j < header.size(); ++j)
	{
		if (boost::regex_match(header[j], pattern1))
		{
			std::string val = boost::regex_replace(header[j], pattern1, "$1");
			if (val.size() != 0)
			{
				col = atoi(val.c_str());
				if (col > nc) nc = col;
			}
		}
	}

	// find col numbers
	for(size_t i = 0; i < ncols; ++i)
	{
		std::ostringstream ossp;
		ossp << ".*[ \t]([0-9]*):" << col_names[i] << ".*";
		boost::regex pattern(ossp.str());
//std::cout << "looking for col " << col_names[i] << " with pattern " << pattern << std::endl;

		bool found = false;
		for(size_t j = 0; j < header.size(); ++j)
		{
			if (boost::regex_match(header[j], pattern))
			{
//std::cout << "found in line : " << header[j] << std::endl;
				std::string val = boost::regex_replace(header[j], pattern, "$1");
				if (val.size() == 0) { std::cerr << "Failed to extract col number for col " << col_names[i] << std::endl; return false; }
				ia(i) = atoi(val.c_str())-1;
//std::cout << "extracted value : " << val << std::endl;
				found = true;
				break;
			}
		}
		if (!found) { std::cerr << "Failed to find col " << col_names[i] << std::endl; return false; }
	}
	file.clear();
	file.seekg(0, file.beg);
	return true;
}


bool read_vec(std::ifstream & file, jblas::vec & v)
{
	std::string line, tmp;
	bool found = false;
	while (!found)
	{
		std::getline(file, line);
		bool res = file.eof();
		if (res) return false;
		if (line.size() < 5 || line[0] == '#') continue; else found = true;
		std::istringstream iss(line);
		for(size_t i = 0; i < v.size(); i++)
		{
			if (iss.eof()) { std::cout << "Not enough elements in line " << line << std::endl; return false; }
			iss >> tmp; v(i) = atof(tmp.c_str());
		}
	}
	return found;
}


jblas::sym_mat build_cov(jblas::vec const & v, jblas::ind_array const & ia_s, jblas::ind_array const & ia_c)
{
	size_t n = ia_s.size();
	jblas::sym_mat P(n,n); P.clear();
	for(size_t j = 0; j < n; ++j) P(j,j) = jmath::sqr(v(ia_s[j]));
	for(size_t i = 1, k = 0; i < n; ++i)
		for(size_t j = 0; j < n-i; ++j, ++k) P(i-1,j+i) = v(ia_c[k]);
JFR_ASSERT(P(1,0) == v(ia_c[0]), "writing in sym_mat does not work " << P << " " << ublas::project(v, ia_c));
	JFR_ASSERT(ublasExtra::det(P) >= -1e-9, "neg det " << ublasExtra::det(P) << " for P " << P << " vec " << v << " ia_s " << ia_s << " ia_c " << ia_c);
	return P;
}

void write_cov(jblas::sym_mat const & P, jblas::vec & v, jblas::ind_array const & ia_s, jblas::ind_array const & ia_c)
{
//std::cout << "write_cov " << P << std::endl;
//std::cout << "write_cov before " << ublas::project(v, ia_s) << " ; " << ublas::project(v, ia_c) << " ; " << v << std::endl;
	JFR_ASSERT(ublasExtra::det(P) >= -1e-9, "neg det " << ublasExtra::det(P) << " for P " << P << " vec " << v << " ia_s " << ia_s << " ia_c " << ia_c);
	size_t n = ia_s.size();
	for(size_t i = 0; i < n; ++i) v(ia_s[i]) = sqrt(P(i,i));
	for(size_t i = 1, k = 0; i < n; ++i)
		for(size_t j = 0; j < n-i; ++j, ++k) v(ia_c[k]) = P(i-1,j+i);
//std::cout << "write_cov after " << ublas::project(v, ia_s) << " ; " << ublas::project(v, ia_c) << " ; " << v << std::endl;
}


void normalize_q(jblas::vec & v, jblas::ind_array & ia_q, jblas::ind_array & ia_sq, jblas::ind_array & ia_cq)
{
	const size_t n = 4;
	jblas::vec q = ublas::project(v, ia_q);
	jblas::sym_mat Q = build_cov(v, ia_sq, ia_cq);

	jblas::vec qn = q; jblas::mat QN_q(n,n);
	ublasExtra::normalize(qn);
	ublasExtra::normalizeJac(q, QN_q);
	jblas::sym_mat QN = ublasExtra::prod_JPJt(Q, QN_q);

	ublas::project(v, ia_q) = qn;
	write_cov(QN, v, ia_sq, ia_cq);
}

class Sep
{
	friend std::ostream& operator<<(std::ostream & s, Sep & sep);
	size_t i, i0;
	char sep;
	public:
		Sep(size_t i0 = 1, char sep = '\t'): i0(i0), sep(sep) { reset(); }
		void reset() { i = i0; }
};
std::ostream& operator<<(std::ostream & s, Sep & sep) { if (sep.i!=sep.i0) s << sep.sep; s << sep.i << ":"; ++sep.i; return s; }
/*
class HeaderField // not working
{
	friend std::ostream& operator<<(std::ostream & s, HeaderField & f);
	size_t i, i0;
	char sep;
	std::string field_;
	public:
		HeaderField(size_t i0 = 1, char sep = '\t'): i0(i0), sep(sep) { reset(); }
		HeaderField & operator()(std::string field) { field_ = field; return *this; }
		void reset() { i = i0; }
};
std::ostream& operator<<(std::ostream & s, HeaderField & f) { if (f.i!=f.i0) s << f.sep; s << f.i << ":" << f.field_; ++f.i; return s; }
*/

int main(int argc, char* const* argv)
{

	if (argc < 4) { std::cerr << "Usage: demo_error <file-ref> <file-eval> <section-length> [s-scale]" << std::endl; return -1; }
	std::ifstream file_ref(argv[1]);
	std::ifstream file_eval(argv[2]);
	double length = atof(argv[3]);
	double s_scale = 0.0;
	bool comp_scale = false;
	if (argc > 4)
	{
		s_scale = atof(argv[4]);
		comp_scale = true;
	}
	srand(jmath::get_srand());

	jblas::vec q0 = quaternion::identQ();

	std::ostringstream oss;
	oss.str(""); oss << argv[2] << ".errors.log";
	std::ofstream file_errors(oss.str().c_str());
	oss.str(""); oss << argv[2] << ".stats.log";
	std::ofstream file_stats(oss.str().c_str());
	oss.str(""); oss << argv[2] << ".diverge.log";
	std::ofstream file_diverge(oss.str().c_str());
	oss.str(""); oss << argv[2] << ".mse.log";
	std::ofstream file_mse(oss.str().c_str());
std::cout << "Find col numbers" << std::endl;
	// first find out column numbers
	jblas::ind_array ia_state_ref(ncols), ia_state_eval(ncols);
	size_t nc_ref, nc_eval;
	if (!find_cols(file_ref, ia_state_ref, nc_ref) || !find_cols(file_eval, ia_state_eval, nc_eval)) return -1;

	jblas::ind_array ia_p = ublasExtra::ia_set(ublas::range(1,1+3));
	jblas::ind_array ia_q = ublasExtra::ia_set(ublas::range(4,4+4));
	jblas::ind_array ia_v = ublasExtra::ia_set(ublas::range(8,8+3));
	jblas::ind_array ia_sp = ublasExtra::ia_set(ublas::range(11,11+3));
	jblas::ind_array ia_sq = ublasExtra::ia_set(ublas::range(14,14+4));
	jblas::ind_array ia_sv = ublasExtra::ia_set(ublas::range(18,18+3));
	jblas::ind_array ia_cp = ublasExtra::ia_set(ublas::range(21,21+3));
	jblas::ind_array ia_cq = ublasExtra::ia_set(ublas::range(24,24+6));
	jblas::ind_array ia_cv = ublasExtra::ia_set(ublas::range(30,30+3));

	size_t ia_t_ref = ia_state_ref[0];
	jblas::ind_array ia_p_ref = ia_state_ref.compose(ia_p);
JFR_ASSERT(ia_p_ref[0] == ia_state_ref[ia_p[0]], "compose in wrong way");
	jblas::ind_array ia_q_ref = ia_state_ref.compose(ia_q);
	jblas::ind_array ia_v_ref = ia_state_ref.compose(ia_v);
	jblas::ind_array ia_sp_ref = ia_state_ref.compose(ia_sp);
	jblas::ind_array ia_sq_ref = ia_state_ref.compose(ia_sq);
	jblas::ind_array ia_sv_ref = ia_state_ref.compose(ia_sv);
	jblas::ind_array ia_cp_ref = ia_state_ref.compose(ia_cp);
	jblas::ind_array ia_cq_ref = ia_state_ref.compose(ia_cq);
	jblas::ind_array ia_cv_ref = ia_state_ref.compose(ia_cv);

	size_t ia_t_eval = ia_state_eval[0];
	jblas::ind_array ia_p_eval = ia_state_eval.compose(ia_p);
	jblas::ind_array ia_q_eval = ia_state_eval.compose(ia_q);
	jblas::ind_array ia_v_eval = ia_state_eval.compose(ia_v);
	jblas::ind_array ia_sp_eval = ia_state_eval.compose(ia_sp);
	jblas::ind_array ia_sq_eval = ia_state_eval.compose(ia_sq);
	jblas::ind_array ia_sv_eval = ia_state_eval.compose(ia_sv);
	jblas::ind_array ia_cp_eval = ia_state_eval.compose(ia_cp);
	jblas::ind_array ia_cq_eval = ia_state_eval.compose(ia_cq);
	jblas::ind_array ia_cv_eval = ia_state_eval.compose(ia_cv);

std::cout << "Sync files, and compute distance and scale" << std::endl;
	// now sync them, and compute distance and scale
	double scale = 1.0;
	double s = 0.0;
	size_t prev_i_s = 0;
	std::vector<Line> lines;
	Line line;
	bool found;
	jblas::vec orient(7); orient.clear(); ublas::subrange(orient,3,7) = quaternion::identQ();

	jblas::vec line_ref(nc_ref), prev_line_ref(nc_ref), int_line_ref(nc_ref);
	jblas::vec line_eval(nc_eval);
	read_vec(file_ref, line_ref);

	while (read_vec(file_eval, line_eval))
	{
		if (line_ref(ia_t_ref) <= line_eval(ia_t_eval))
		{
			prev_line_ref = line_ref;

			found = false;
			while (read_vec(file_ref, line_ref))
				if (line_ref(ia_t_ref) > line_eval(ia_t_eval)) { found = true; break; }
			if (!found) break;
		}

		double coeff = (line_eval(ia_t_eval)-prev_line_ref(ia_t_ref)) / (line_ref(ia_t_ref)-prev_line_ref(ia_t_ref));
		line.t = line_eval(ia_t_eval);
		line.s = s;
		line.ref = prev_line_ref + (line_ref-prev_line_ref) * coeff;
//std::cout << "prev " << prev_line_ref << " ; current " << line_ref << " ; coeff " << coeff << " ; new " << line.ref << std::endl;
		normalize_q(line.ref, ia_q_ref, ia_sq_ref, ia_cq_ref);
		normalize_q(line_eval, ia_q_eval, ia_sq_eval, ia_cq_eval); // FIXME can be removed when fixed in logs

		// align roll/pitch
		if (lines.size() == 0)
		{
			jblas::vec euler_eval = quaternion::q2e(ublas::project(line_eval, ia_q_eval));
			jblas::vec euler_ref = quaternion::q2e(ublas::project(line_ref, ia_q_ref));
			jblas::vec euler(3); euler.clear();
			if (euler_eval(0) == 0.0 && euler_eval(1) == 0.0 && euler_eval(2) == 0.0)
				ublas::subrange(orient,3,7) = quaternion::e2q(euler_ref);
		}
		jblas::vec newf_eval(7), f_eval = ublas::project(line_eval, ublasExtra::ia_union(ia_p_eval,ia_q_eval));
		jblas::sym_mat newF_eval(7,7), F_eval(7,7); F_eval.clear();
		ublas::subrange(F_eval,0,3,0,3) = build_cov(line_eval, ia_sp_eval, ia_cp_eval);
		ublas::subrange(F_eval,3,7,3,7) = build_cov(line_eval, ia_sq_eval, ia_cq_eval);
		jblas::mat F_f1(7,7), F_f2(7,7);
		//quaternion::qProd(orient, q_eval, newq_eval, Q_q1, Q_q2);
		quaternion::composeFrames(orient, f_eval, newf_eval, F_f1, F_f2);
		newF_eval = ublasExtra::prod_JPJt(F_eval, F_f1);

//std::cout << "after orient, replace " << f_eval <<  " ; " << F_eval << "  with  " << newf_eval << " ; " << newF_eval << std::endl;
		ublas::project(line_eval, ublasExtra::ia_union(ia_p_eval,ia_q_eval)) = newf_eval;
		write_cov(ublas::subrange(newF_eval,0,3,0,3), line_eval, ia_sp_eval, ia_cp_eval);
		write_cov(ublas::subrange(newF_eval,3,7,3,7), line_eval, ia_sq_eval, ia_cq_eval);
		JFR_ASSERT(ublasExtra::det(newF_eval) >= -1e-9, "neg det " << ublasExtra::det(newF_eval) << " for newF_eval " << newF_eval);

		line.eval = line_eval;
		lines.push_back(line);

jblas::sym_mat tmp; // just to check nan
tmp = build_cov(line.ref, ia_sp_ref, ia_cp_ref);
tmp = build_cov(line.ref, ia_sq_ref, ia_cq_ref);
tmp = build_cov(line.ref, ia_sv_ref, ia_cv_ref);
tmp = build_cov(line.eval, ia_sp_eval, ia_cp_eval);
tmp = build_cov(line.eval, ia_sq_eval, ia_cq_eval);
tmp = build_cov(line.eval, ia_sv_eval, ia_cv_eval);

		// compute s
		double norm_s = ublas::norm_2(ublas::project(line.ref, ia_p_ref)-ublas::project(lines[prev_i_s].ref, ia_p_ref));
		if (norm_s > 0.1) // 10cm
		{
			for(size_t i = prev_i_s+1; i < lines.size(); ++i)
				lines[i].s += norm_s * (lines[i].t - lines[prev_i_s].t) / (line.t - lines[prev_i_s].t);
			prev_i_s = lines.size()-1;
			s = lines.back().s;
		}

		// compute scale
		if (comp_scale && line.s > s_scale)
		{
			scale = ublas::norm_2(ublas::project(line.eval, ia_p_eval)-ublas::project(lines[0].eval, ia_p_eval))
						/ ublas::norm_2(ublas::project(line.ref, ia_p_ref)-ublas::project(lines[0].ref, ia_p_ref));
			comp_scale = false;
std::cout << "initialize scale to " << scale << "(i'm in " << ublas::project(line.eval, ia_p_eval) << ", should be in "
	<< ublas::project(line.ref, ia_p_ref) << ")" << std::endl;
std::cout << "we are at time " << std::setprecision(16) << line.t << " s " << s_scale << std::endl;
		}
	}

std::cout << "Output the error file (" << lines.size() << " lines)" << std::endl;
	// now output the error file
	Sep p;
	file_errors << "#";
	file_errors << p << "time" << p << "s";
	file_errors << p << "ref_x" << p << "ref_y" << p << "ref_z";
	file_errors << p << "ref_yaw" << p << "ref_pitch" << p << "ref_roll";
	file_errors << p << "ref_velx" << p << "ref_vely" << p << "ref_velz" << p << "ref_veln";
	file_errors << p << "ref_sig_x" << p << "ref_sig_y" << p << "ref_sig_z";
	file_errors << p << "ref_sig_yaw" << p << "ref_sig_pitch" << p << "ref_sig_roll";
	file_errors << p << "ref_sig_velx" << p << "ref_sig_vely" << p << "ref_sig_velz" << p << "ref_sig_veln";
	file_errors << p << "eval_x" << p << "eval_y" << p << "eval_z";
	file_errors << p << "eval_yaw" << p << "eval_pitch" << p << "eval_roll";
	file_errors << p << "eval_velx" << p << "eval_vely" << p << "eval_velz" << p << "eval_veln";
	file_errors << p << "eval_sig_x" << p << "eval_sig_y" << p << "eval_sig_z";
	file_errors << p << "eval_sig_yaw" << p << "eval_sig_pitch" << p << "eval_sig_roll";
	file_errors << p << "eval_sig_velx" << p << "eval_sig_vely" << p << "eval_sig_velz" << p << "eval_sig_veln";
	file_errors << p << "err_x" << p << "err_y" << p << "err_z" << p << "err_xy";
	file_errors << p << "err_yaw" << p << "err_pitch" << p << "err_roll";
	file_errors << p << "err_velx" << p << "err_vely" << p << "err_velz" << p << "err_veln";
	file_errors << p << "err_sig_x" << p << "err_sig_y" << p << "err_sig_z" << p << "err_sig_xy";
	file_errors << p << "err_sig_yaw" << p << "err_sig_pitch" << p << "err_sig_roll";
	file_errors << p << "err_sig_velx" << p << "err_sig_vely" << p << "err_sig_velz" << p << "err_sig_veln";
	file_errors << p << "nees_x" << p << "nees_y" << p << "nees_z" << p << "nees_xy";
	file_errors << p << "nees_yaw" << p << "nees_pitch" << p << "nees_roll";
	file_errors << p << "nees_velx" << p << "nees_vely" << p << "nees_velz" << p << "nees_veln";
	file_errors << std::endl;
	double mse_p = 0.0, mse_pn = 0.0, mse_q = 0.0, mse_qn = 0.0, mse_yawn = 0.0;
	for(size_t i = 0; i < lines.size(); ++i)
	{
//std::cout << "process line " << i << std::endl;
		// scale
		ublas::project(lines[i].eval, ia_p_eval) /= scale;
		ublas::project(lines[i].eval, ia_sp_eval) /= scale;
		ublas::project(lines[i].eval, ia_cp_eval) /= scale*scale;
		ublas::project(lines[i].eval, ia_v_eval) /= scale;
		ublas::project(lines[i].eval, ia_sv_eval) /= scale;
		ublas::project(lines[i].eval, ia_cv_eval) /= scale*scale;

		// output common
		file_errors << std::setprecision(16) << lines[i].t << std::setprecision(6) << "\t" << lines[i].s << "\t";

		// build ref
		jblas::sym_mat Pq_ref = build_cov(lines[i].ref, ia_sq_ref, ia_cq_ref), Pe_ref(3,3);
		jblas::vec q_ref = ublas::project(lines[i].ref, ia_q_ref), e_ref(3);
		quaternion::q2e(q_ref, Pq_ref, e_ref, Pe_ref);

		jblas::mat VN_v(1,3);
		jblas::sym_mat Pv_ref = build_cov(lines[i].ref, ia_sv_ref, ia_cv_ref);
		double vn_ref = ublasExtra::norm_2(ublas::project(lines[i].ref,ia_v_ref), VN_v);
		double Pvn_ref = ublasExtra::prod_JPJt(Pv_ref, VN_v)(0,0);

		// build eval
		jblas::sym_mat Pq_eval = build_cov(lines[i].eval, ia_sq_eval, ia_cq_eval), Pe_eval(3,3);
		jblas::vec q_eval = ublas::project(lines[i].eval, ia_q_eval), e_eval(3);
		quaternion::q2e(q_eval, Pq_eval, e_eval, Pe_eval);

		jblas::sym_mat Pv_eval = build_cov(lines[i].eval, ia_sv_eval, ia_cv_eval);
		double vn_eval = ublasExtra::norm_2(ublas::project(lines[i].eval,ia_v_eval), VN_v);
		double Pvn_eval = ublasExtra::prod_JPJt(Pv_eval, VN_v)(0,0);

		// output ref
		for(int j = 0; j < 3; ++j) file_errors << lines[i].ref(ia_p_ref[j]) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << e_ref(2-j) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << lines[i].ref(ia_v_ref[j]) << "\t";
		file_errors << vn_ref << "\t";

		for(int j = 0; j < 3; ++j) file_errors << lines[i].ref(ia_sp_ref[j]) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << sqrt(Pe_ref(2-j,2-j)) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << lines[i].ref(ia_sv_ref[j]) << "\t";
		file_errors << sqrt(Pvn_ref) << "\t";

		// output eval
		for(int j = 0; j < 3; ++j) file_errors << lines[i].eval(ia_p_eval[j]) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << e_eval(2-j) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << lines[i].eval(ia_v_eval[j]) << "\t";
		file_errors << vn_eval << "\t";

		for(int j = 0; j < 3; ++j) file_errors << lines[i].eval(ia_sp_eval[j]) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << sqrt(Pe_eval(2-j,2-j)) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << lines[i].eval(ia_sv_eval[j]) << "\t";
		file_errors << sqrt(Pvn_eval) << "\t";

		// build error
		jblas::mat PN_p(1,2);
		jblas::ind_array ia_xy = ublasExtra::ia_set(ublas::range(0,2));
		jblas::vec p = ublas::project(lines[i].eval,ia_p_eval) - ublas::project(lines[i].ref,ia_p_ref);
		double xy = ublasExtra::norm_2(ublas::project(p,ia_xy), PN_p);
		jblas::sym_mat Pp = build_cov(lines[i].ref, ia_sp_ref, ia_cp_ref) + build_cov(lines[i].eval, ia_sp_eval, ia_cp_eval);
		jblas::sym_mat Pxy = ublas::project(Pp, ia_xy, ia_xy);
		double Pxyn = ublasExtra::prod_JPJt(Pxy, PN_p)(0,0);
		jblas::sym_mat Pv = Pv_ref + Pv_eval;

		for(int j = 0; j < 3; ++j) file_errors << lines[i].eval(ia_p_eval[j])-lines[i].ref(ia_p_ref[j]) << "\t"; // x,y,z
		file_errors << xy << "\t"; // xy
		for(int j = 0; j < 3; ++j) { double err = e_eval(2-j)-e_ref(2-j); if (err > M_PI) err-=2*M_PI; if (err<-M_PI) err+=2*M_PI; file_errors << err << "\t"; } //w,p,r
		for(int j = 0; j < 3; ++j) file_errors << lines[i].eval(ia_v_eval[j])-lines[i].ref(ia_v_ref[j]) << "\t"; //vx,vy,vz
		file_errors << vn_eval-vn_ref << "\t"; // v

		for(int j = 0; j < 3; ++j) file_errors << sqrt(jmath::sqr(lines[i].eval(ia_sp_eval[j]))+jmath::sqr(lines[i].ref(ia_sp_ref[j]))) << "\t";
		file_errors << sqrt(Pxyn) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << sqrt(Pe_eval(2-j,2-j)+Pe_ref(2-j,2-j)) << "\t";
		for(int j = 0; j < 3; ++j) file_errors << sqrt(jmath::sqr(lines[i].eval(ia_sv_eval[j]))+jmath::sqr(lines[i].ref(ia_sv_ref[j]))) << "\t";
		file_errors << sqrt(Pvn_eval+Pvn_ref) << "\t";

		jblas::vec frame_ref(7);
		ublas::subrange(frame_ref,0,3) = ublas::project(lines[i].ref,ia_p_ref);
		ublas::subrange(frame_ref,3,7) = ublas::project(lines[i].ref,ia_q_ref);
		jblas::vec frame_eval(7);
		ublas::subrange(frame_eval,0,3) = ublas::project(lines[i].eval,ia_p_eval);
		ublas::subrange(frame_eval,3,7) = ublas::project(lines[i].eval,ia_q_eval);
		jblas::vec frame = quaternion::frameIncrement(frame_ref, frame_eval);
		mse_p += jmath::sqr(ublas::norm_2(ublas::subrange(frame, 0, 3)));
		double scal = ublas::inner_prod(ublas::subrange(frame, 3, 7), q0); if (scal > 1) scal = 1.0; // because of numerical errors
		mse_q += jmath::sqr(acos(2*scal-1));

		// build nees
		double err, var;
		for(int j = 0; j < 3; ++j) { var = jmath::sqr(lines[i].eval(ia_sp_eval[j]))+jmath::sqr(lines[i].ref(ia_sp_ref[j])); file_errors << (var < 1e-15 ? 0.0 : jmath::sqr(lines[i].eval(ia_p_eval[j])-lines[i].ref(ia_p_ref[j]))/var) << "\t"; } // x,y,z
		file_errors << (ublasExtra::det(Pxy) < 1e-15 ? 0.0 : ublasExtra::prod_xt_P_x(Pxy, ublas::project(p,ia_xy))) << "\t"; // xy
		for(int j = 0; j < 3; ++j) { var = Pe_eval(2-j,2-j)+Pe_ref(2-j,2-j); err = e_eval(2-j)-e_ref(2-j); if (err > M_PI) err-=2*M_PI; if (err<-M_PI) err+=2*M_PI; double n = (var < 1e-15 ? 0.0 : (err*err)/var); file_errors << n << "\t"; if (j==0) mse_yawn += n; } //w,p,r
		for(int j = 0; j < 3; ++j) { var = jmath::sqr(lines[i].eval(ia_sv_eval[j]))+jmath::sqr(lines[i].ref(ia_sv_ref[j])); file_errors << (var < 1e-15 ? 0.0 : jmath::sqr(lines[i].eval(ia_v_eval[j])-lines[i].ref(ia_v_ref[j]))/var) << "\t"; } //vx,vy,vz
		file_errors << (ublasExtra::det(Pv) < 1e-15 ? 0.0 : ublasExtra::prod_xt_P_x(Pv, ublas::project(lines[i].eval,ia_v_eval)-ublas::project(lines[i].ref,ia_v_ref)))  << "\t"; // v

		mse_pn += (ublasExtra::det(Pp) < 1e-20 ? 0.0 : ublasExtra::prod_xt_P_x(Pp, p));
		mse_qn += (ublasExtra::det(Pq_ref+Pq_eval) < 1e-50 ? 0.0 : ublasExtra::prod_xt_P_x(Pq_ref+Pq_eval, q_eval-q_ref));

		file_errors << std::endl;
	}
	file_errors.close();
	mse_p = (mse_p / lines.size());
	mse_q = (mse_q / lines.size());
	mse_pn = (mse_pn / lines.size());
	mse_qn = (mse_qn / lines.size());
	mse_yawn = (mse_qn / lines.size());
	file_mse << mse_p << "\t" << mse_q << "\t" << mse_pn << "\t" << mse_qn << "\t" << mse_yawn << std::endl;
	file_mse.close();

std::cout << "Compute stats" << std::endl;
	// now compute stats
	if (lines.size() == 0) { std::cerr << "could not read any line" << std::endl; return -1; }
	//file_stats << "#1:file\t2:index\t3:s0\t4:s1\t5:pos_err\t6:ori_err\t7:pos_nees\t8:ori_nees" << std::endl;
	file_stats << "## " << "length: " << length << std::endl;
	p.reset(); file_stats << "#" << p << "file" << p << "index" << p << "s0" << p << "s1" << p << "pos_err" << p << "ori_err" << p << "pos_nees" << p << "ori_nees" << std::endl;
	//HeaderField f; file_stats << "#" << f("file") << f("index") << f("s0") << f("s1") << f("pos_err") << f("ori_err") << f("pos_nees") << f("ori_nees") << std::endl;
	double end_s, prev_s = lines.back().s * (rand()%1000)/1000.;
	size_t prev_k, k = std::floor(prev_s / length);
	prev_s -= k*length;
	for(k = 0; k < lines.size(); ++k) if (lines[k].s >= prev_s) break;

	bool diverge = false;
	for(int i = 0; ; ++i)
	{
		end_s = prev_s + length;
std::cout << "  For segment " << prev_s << " to " << end_s << std::endl;
		prev_k = k;
		for(k = 0; k < lines.size(); ++k) if (lines[k].s >= end_s) break;
		if (k >= lines.size()) break;

//std::cout << "line " << prev_k << " ref " << lines[prev_k].ref << std::endl;
//std::cout << "line " << k << " ref " << lines[k].ref << std::endl;
//std::cout << "line " << prev_k << " eval " << lines[prev_k].eval << std::endl;
//std::cout << "line " << k << " eval " << lines[k].eval << std::endl;
		// build frames
		jblas::sym_mat frameAB_ref(14,14); frameAB_ref.clear();
		jblas::vec frameA_ref(7); frameA_ref.clear();
		ublas::subrange(frameA_ref,0,3) = ublas::project(lines[prev_k].ref,ia_p_ref);
		ublas::subrange(frameA_ref,3,7) = ublas::project(lines[prev_k].ref,ia_q_ref);
		ublas::subrange(frameAB_ref,0,3,0,3) = build_cov(lines[prev_k].ref, ia_sp_ref, ia_cp_ref);
		ublas::subrange(frameAB_ref,3,7,3,7) = build_cov(lines[prev_k].ref, ia_sq_ref, ia_cq_ref);
		jblas::vec frameB_ref(7); frameB_ref.clear();
		ublas::subrange(frameB_ref,0,3) = ublas::project(lines[k].ref,ia_p_ref);
		ublas::subrange(frameB_ref,3,7) = ublas::project(lines[k].ref,ia_q_ref);
		ublas::subrange(frameAB_ref,7,10,7,10) = build_cov(lines[k].ref, ia_sp_ref, ia_cp_ref);
		ublas::subrange(frameAB_ref,10,14,10,14) = build_cov(lines[k].ref, ia_sq_ref, ia_cq_ref);

		jblas::sym_mat frameAB_eval(14,14); frameAB_eval.clear();
		jblas::vec frameA_eval(7); frameA_eval.clear();
		ublas::subrange(frameA_eval,0,3) = ublas::project(lines[prev_k].eval,ia_p_eval);
		ublas::subrange(frameA_eval,3,7) = ublas::project(lines[prev_k].eval,ia_q_eval);
		ublas::subrange(frameAB_eval,0,3,0,3) = build_cov(lines[prev_k].eval, ia_sp_eval, ia_cp_eval);
		ublas::subrange(frameAB_eval,3,7,3,7) = build_cov(lines[prev_k].eval, ia_sq_eval, ia_cq_eval);
		jblas::vec frameB_eval(7); frameB_eval.clear();
		ublas::subrange(frameB_eval,0,3) = ublas::project(lines[k].eval,ia_p_eval);
		ublas::subrange(frameB_eval,3,7) = ublas::project(lines[k].eval,ia_q_eval);
		ublas::subrange(frameAB_eval,7,10,7,10) = build_cov(lines[k].eval, ia_sp_eval, ia_cp_eval);
		ublas::subrange(frameAB_eval,10,14,10,14) = build_cov(lines[k].eval, ia_sq_eval, ia_cq_eval);
//		ublas::subrange(frameAB_eval,7,14,0,7) = ublas::subrange(frameAB_eval,0,7,0,7);
//JFR_ASSERT(frameAB_eval(0,7) == frameAB_eval(0,0), "error sym matrix2");

//std::cout << "frameAB_eval " << frameAB_eval << std::endl;

		// compose frames
		jblas::mat J1(7,7), J2(7,7), J12(7,14);
		jblas::vec frame_ref(7);
		jblas::sym_mat frameP_ref(7,7);
		quaternion::frameIncrement(frameA_ref, frameB_ref, frame_ref, J1, J2);
		ublas::subrange(J12,0,7,0,7) = J1;
		ublas::subrange(J12,0,7,7,14) = J2;
		frameP_ref = ublasExtra::prod_JPJt(frameAB_ref, J12);

		jblas::vec frame_eval(7);
		jblas::sym_mat frameP_eval(7,7);
		quaternion::frameIncrement(frameA_eval, frameB_eval, frame_eval, J1, J2);
		ublas::subrange(J12,0,7,0,7) = J1;
		ublas::subrange(J12,0,7,7,14) = J2;
		frameP_eval = ublasExtra::prod_JPJt(frameAB_eval, J12);

		// error
		jblas::vec frame(7);
		jblas::sym_mat frameP(7,7);
		quaternion::frameIncrement(frame_ref, frame_eval, frame, J1, J2);
		frameP = ublasExtra::prod_JPJt(frameP_ref, J1) + ublasExtra::prod_JPJt(frameP_eval, J2);

//std::cout << "frameA_ref " << frameA_ref << " P " << ublas::subrange(frameAB_ref,0,3,0,3) << " " << ublas::subrange(frameAB_ref,3,7,3,7) << std::endl;
//std::cout << "frameB_ref " << frameB_ref << " P " << ublas::subrange(frameAB_ref,7,10,7,10) << " " << ublas::subrange(frameAB_ref,3,7,3,7) << std::endl;
//std::cout << "frameA_eval " << frameA_eval << " P " << ublas::subrange(frameAB_eval,0,3,0,3) << " " << ublas::subrange(frameAB_eval,3,7,3,7) << std::endl;
//std::cout << "frameB_eval " << frameB_eval << " P " << ublas::subrange(frameAB_eval,7,10,7,10) << " " << ublas::subrange(frameAB_eval,3,7,3,7) << std::endl;
//std::cout << "frame_ref " << frame_ref << " P " << ublas::subrange(frameP_ref,0,3,0,3) << " " << ublas::subrange(frameP_ref,3,7,3,7) << std::endl;
//std::cout << "frame_eval " << frame_eval << " P " << ublas::subrange(frameP_eval,0,3,0,3) << " " << ublas::subrange(frameP_eval,3,7,3,7) << std::endl;
//std::cout << "frame " << frame << " P " << ublas::subrange(frameP,0,3,0,3) << " " << ublas::subrange(frameP,3,7,3,7) << std::endl;

JFR_ASSERT(ublasExtra::det(frameP_ref) >= -1e-9, "neg det " << ublasExtra::det(frameP_ref) << " for frameP_ref " << frameP_ref);
JFR_ASSERT(ublasExtra::det(frameP_eval) >= -1e-9, "neg det " << ublasExtra::det(frameP_eval) << " for frameP_eval " << frameP_eval);
		// check diverged
		if (ublas::norm_2(ublas::subrange(frame, 0, 3)) > length * 0.5) { diverge = true; break; }

		// write
		file_stats << argv[2] << "\t" << i << "\t" << prev_s << "\t" << end_s << "\t";

		file_stats << ublas::norm_2(ublas::subrange(frame, 0, 3)) << "\t";
		double scal = ublas::inner_prod(ublas::subrange(frame, 3, 7), q0); if (scal > 1) scal = 1.0; // because of numerical errors
		file_stats << acos(2*scal-1) << "\t";

		// nees
//std::cout << "pos x " << ublas::subrange(frame, 0, 3) << " detP " << ublasExtra::det(ublas::subrange(frameP,0,3,0,3)) << " P " << ublas::subrange(frameP,0,3,0,3) << std::endl;
//std::cout << "ori x " << ublas::subrange(frame, 3, 7)-q0 << " detP " << ublasExtra::det(ublas::subrange(frameP,3,7,3,7)) << " P " << ublas::subrange(frameP,3,7,3,7) << std::endl;
		file_stats << (ublasExtra::det(ublas::subrange(frameP,0,3,0,3)) < 1e-20 ? 0.0 : ublasExtra::prod_xt_P_x(ublas::subrange(frameP,0,3,0,3), ublas::subrange(frame, 0, 3))) << "\t";
		file_stats << (ublasExtra::det(ublas::subrange(frameP,3,7,3,7)) < 1e-50 ? 0.0 : ublasExtra::prod_xt_P_x(ublas::subrange(frameP,3,7,3,7), ublas::subrange(frame, 3, 7)-q0)) << "\t";
		file_stats << std::endl;

		prev_s = end_s;
	}
	file_stats.close();

	file_diverge << argv[2] << "\t" << diverge << std::endl;
	file_diverge.close();

	return 0;
}

