/**
 * \file hardwareSensorCameraFirewire.hpp
 *
 * Header file for getting data from a firewire camera
 *
 * \date 18/06/2010
 * \author croussil
 *
 * \ingroup rtslam
 */

#ifndef HARDWARE_SENSOR_CAMERA_HPP_
#define HARDWARE_SENSOR_CAMERA_HPP_

#include <image/Image.hpp>
#include <kernel/threads.hpp>

#include <boost/thread.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>

#include "rtslam/hardwareSensorAbstract.hpp"
#include "rtslam/rawImage.hpp"


namespace jafar {
namespace rtslam {
namespace hardware {

/**
This class allows to get images from firewire with non blocking procedure,
using triple-buffering.
*/
class HardwareSensorCamera: public HardwareSensorExteroAbstract
{
	protected:
		class LoggableImage: public kernel::Loggable
		{
		 private:
			std::string dump_path;
			rawimage_ptr_t image;
			int index;
			size_t id;
			std::fstream &fb;

		 public:
			LoggableImage(std::string const & dump_path, rawimage_ptr_t const & image, int index, size_t id, std::fstream &fb):
				dump_path(dump_path), image(image), index(index), id(id), fb(fb) {}
			virtual void log()
			{
				std::ostringstream oss; oss << dump_path << "/image_";
				if (id < 1000) oss << id << "_";
				oss << std::setw(7) << std::setfill('0') << index;
				image->img->save(oss.str() + std::string(".pgm"));
				std::fstream f; f.open((oss.str() + std::string(".time")).c_str(), std::ios_base::out);
				f << std::setprecision(19) << image->timestamp << std::endl; f.close();
				fb.write((char*)&(index), sizeof(int));
				fb.write((char*)&(image->timestamp), sizeof(double));
			}
		};

	protected:
		std::vector<IplImage*> bufferImage;
		std::vector<rawimage_ptr_t> bufferSpecPtr;
		std::list<rawimage_ptr_t> bufferSave;
		unsigned index;
		int found_first; /// 0 = not found, 1 = found pgm, 2 = found png
		
		double realFreq;
		double last_timestamp;
		std::string dump_path;
		
		boost::thread *preloadTask_thread;
		virtual void preloadTask(void) = 0;
		void preloadTaskOffline(void);
		kernel::LoggerTask *loggerTask;

		int cam_id;
		int filter_div, filter_mod;

		void init(Mode mode, std::string dump_path, cv::Size imgSize);
		virtual void start();
		virtual void stop();
		virtual bool join(int timed_ms = -1);

	public:
		HardwareSensorCamera(kernel::VariableCondition<int> *condition, int cam_id, cv::Size imgSize, std::string dump_path = ".");
		HardwareSensorCamera(kernel::VariableCondition<int> *condition, Mode mode, int cam_id, int bufferSize, kernel::LoggerTask *loggerTask);

		virtual double getLastTimestamp() { boost::unique_lock<boost::mutex> l(mutex_data); return last_timestamp; }
		double getFreq() { return realFreq; }
		inline int id() const { return cam_id; }
		void setFilter(int div, int mod)
			{ filter_div = div; filter_mod = mod; }
};


}}}

#endif
