/**
 * \file robotConstantMotion.hpp
 *
 * Header file for the robot with constant motion model (position/velocity/acceleration/jerk).
 *
 * \date 17/05/2013
 * \author croussil
 *
 * \ingroup rtslam
 */

#ifndef ROBOTCONSTANTMOTIONMODEL_HPP_
#define ROBOTCONSTANTMOTIONMODEL_HPP_

#include "rtslam/robotAbstract.hpp"

/**
Considers that perturbation is not random but deterministic.
This makes sense because the perturbation is n+1 derivative, which is mostly biased
at the scale of dt.
*/
#define PERT_DETERMINISTIC 0


namespace jafar {
	namespace rtslam {


		/**
		 * Constant velocity model robot class.
		 * The SLAM origin is the camera origin.
		 *
		 * \author croussil
		 *
		 * This class implements a rigid frame in 3D moving with a constant
		 * position/velocity/acceleration/jerk motion model. This model is the following:\n
		 * - p += v*dt <-- position\n
		 * - q = q**(w*dt) <-- orientation quaternion\n
		 * - v += vi <-- linear velocity\n
		 * - w += wi <-- angular velocity
		 *
		 * where \a vi and \a wi are linear and angular velocity random impulses,
		 * and ** is the quaternion product.
		 *
		 * This model is embedded into the system variables as follows:
		 * - the prediction function is x+ = f(x,u), implemented with method move().
		 * - the state vector is x = state.x = [p q v w]
		 * - the state covariance is in the map, P = state.P = (map.filter).P(state.ia, state.ia)
		 * - the control vector is u = control.x = [vi wi]
		 * - the control covariances matrix is U = control.P
		 * - the Jacobians of f(x,u) provided by move() are XNEW_x and XNEW_control.
		 * - the perturbation covariance is obtained with the method computeStatePerturbation(), as follows:
		 * 		- Q = XNEW_control * control.P * XNEW_control'\n
		 * .
		 *
		 * \sa Explore the comments in file robotConstantVelocity.cpp for full algebraic details.
		 *
		 * \ingroup rtslam
		 */
		template<unsigned LIN_ORDER, unsigned ANG_ORDER>
		class RobotConstantMotionModel: public RobotAbstract {
			private:
				ublas::range iaPos, iaOriQuat, iaAbsVel, iaAngVel, iaAbsAcc, iaAngAcc, iaAbsJerk, iaAngJerk;
			public:
				/**
				 * Remote constructor from remote map.
				 * \param _map the remote map
				 */
				RobotConstantMotionModel(const map_ptr_t & _mapPtr);

				~RobotConstantMotionModel(void) {}

				virtual std::string typeName() const {
					return "Constant-motion-model of order " + std::string(1,'0'+LIN_ORDER) + "/" + std::string(1,'0'+ANG_ORDER);
				}


				/**
				 * Move one step ahead.
				 *
				 * This function predicts the robot state one step of length \a dt ahead in time,
				 * according to the control input \a control.x and the time interval \a control.dt.
				 *
				 * \param _x the current state vecto
				 * \param _p the perturbation vector
				 * \param _dt the sampling time
				 * \param _xnew the new state vector
				 * \param _XNEW_x the Jacobian of xnew wrt x
				 * \param _XNEW_pert the Jacobian of xnew wrt p
				 */
				void move_func(const vec & _x, const vec & _u, const vec & _n, const double _dt, vec & _xnew, mat & _XNEW_x,
						mat & _XNEW_pert, unsigned tempSet = 0) const;

				void computePertJacobian();

				static size_t size() {
					return 7+LIN_ORDER*3+ANG_ORDER*3;
				}
				static size_t size_control() {
					return 0;
				}
				static size_t size_perturbation() {
					return 6;
				}

				virtual size_t mySize() {return size();}
				virtual size_t mySize_control() {return size_control();}
				virtual size_t mySize_perturbation() {return size_perturbation();}

				void setVelocityStd(double velLinStd, double velAngStd)
				{
					double factor = 1.;
					unsigned start = pose.size();
					for(size_t j = 0; j < LIN_ORDER; ++j)
					{
						if (j < LIN_ORDER)
						{
							for (size_t i = start; i < start + 3; ++i)
								state.P(i,i) = velLinStd*velLinStd*factor;
							start += 3;
						}
						if (j < ANG_ORDER)
						{
							for (size_t i = start; i < start + 3; ++i)
								state.P(i,i) = velAngStd*velAngStd*factor;
							start += 3;
						}
						factor *= 100;
					}
				}

				virtual void writeLogHeader(kernel::DataLogger& log) const;
				virtual void writeLogData(kernel::DataLogger& log) const;
				
			protected:
				/**
				 * Split state vector.
				 *
				 * Extracts \a p, \a q, \a v and \a w from the state vector, \a x = [\a p, \a q, \a v, \a w].
				 * \param x the state vector
				 * \param p the position
				 * \param q the quaternion
				 * \param v the linear velocity
				 * \param w the angular velocity
				 */
				template<class Vx, class Vp, class Vq, class Vv, class Vw,
								class Vva, class Vwa, class Vvj, class Vwj>
				inline void splitState(const Vx x, Vp & p, Vq & q, Vv & v, Vw & w,
					Vva & va, Vwa & wa, Vvj & vj, Vwj & wj) const {
					switch (LIN_ORDER)
					{
						case 3: vj = ublas::project(x, iaAbsJerk);
						case 2: va = ublas::project(x, iaAbsAcc);
						case 1: v = ublas::project(x, iaAbsVel);
						case 0: p = ublas::project(x, iaPos);
					}
					switch (ANG_ORDER)
					{
						case 3: wj = ublas::project(x, iaAngJerk);
						case 2: wa = ublas::project(x, iaAngAcc);
						case 1: w = ublas::project(x, iaAngVel);
						case 0: q = ublas::project(x, iaOriQuat);
					}
				}


				/**
				 * Compose state vector.
				 *
				 * Composes the state vector with \a p, \a q, \a v and \a w, \a x = [\a p, \a q, \a v, \a w].
				 * \param p the position
				 * \param q the quaternion
				 * \param v the linear velocity
				 * \param w the angular velocity
				 * \param x the state vector
				 */
				template<class Vp, class Vq, class Vv, class Vw,
								class Vva, class Vwa, class Vvj, class Vwj, class Vx>
				inline void unsplitState(const Vp & p, const Vq & q, const Vv & v, const Vw & w,
					const Vva & va, const Vwa & wa, const Vvj & vj, const Vwj & wj, Vx & x) const {
					switch (LIN_ORDER)
					{
						case 3: ublas::project(x, iaAbsJerk) = vj;
						case 2: ublas::project(x, iaAbsAcc) = va;
						case 1: ublas::project(x, iaAbsVel) = v;
						case 0: ublas::project(x, iaPos) = p;
					}
					switch (ANG_ORDER)
					{
						case 3: ublas::project(x, iaAngJerk) = wj;
						case 2: ublas::project(x, iaAngAcc) = wa;
						case 1: ublas::project(x, iaAngVel) = w;
						case 0: ublas::project(x, iaOriQuat) = q;
					}
				}



				/**
				 * Split control vector.
				 *
				 * Extracts impulses \a vi and \a wi from the control vector.
				 * \param vi the linear impulse.
				 * \param wi the angular impulse.
				 */
				template<class Vu, class V>
				inline void splitControl(const Vu & u, V & vi, V & wi) const {
					vi = project(u, ublas::range(0, 3));
					wi = project(u, ublas::range(3, 6));
				}

			private:
				// temporary matrices to speed up Jacobian computation
				struct TempVariables
				{
					mat43 QWDT_wdt; ///< 	Temporary Jacobian matrix
					mat44 QNEW_qwdt; ///<   Temporary Jacobian matrix
					mat43 QNEW_wdt; ///<    Temporary Jacobian matrix
					mat44 QNEW_q; ///<      Temporary Jacobian matrix
					mat44 QNORM_qnew; ///<	Temporary Jacobian matrix
				};
				mutable TempVariables tempvars0, tempvars1;

		};

	}
}

#endif /* ROBOTCONSTANTMOTIONMODEL_HPP_ */
