/**
 * \file sensorAbstract.hpp
 *
 * Header file for the abstract sensor.
 *
 * \date 28/01/2010
 * \author jsola
 *
 * \ingroup rtslam
 */

#ifndef SENSORABSTRACT_H_
#define SENSORABSTRACT_H_

/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

#include <boost/assign.hpp>
#include <boost/smart_ptr.hpp>

#include "kernel/IdFactory.hpp"
#include "jmath/jblas.hpp"
#include "rtslam/rtSlam.hpp"
//include parents
#include "rtslam/parents.hpp"
#include "rtslam/mapAbstract.hpp"
#include "rtslam/mapObject.hpp"
#include "rtslam/robotAbstract.hpp"
#include "rtslam/hardwareSensorAbstract.hpp"

namespace jafar {
	namespace rtslam {

		// Forward declarations of children
		class ObservationAbstract;
		class DataManagerAbstract;

		/** 
			Base class for all types of sensors.
			\ingroup rtslam
		*/
		class SensorAbstract: public MapObject,
		                      public ChildOf<RobotAbstract>,
													public boost::enable_shared_from_this<SensorAbstract>,
													public kernel::DataLoggable
		{
			private:
				// define the function linkToParentRobot().
				ENABLE_LINK_TO_PARENT(RobotAbstract,Robot,SensorAbstract);
				// define the functions robotPtr() and robot().
				ENABLE_ACCESS_TO_PARENT(RobotAbstract,robot);
		
			protected:
				bool integrate_all;
				bool use_for_init; ///< use this sensor to init the state, so needs to process it before those that are not used to init the state
				bool need_init; ///< needs a few seconds of readings to initialize
				bool enabled; /// whether the sensor is enabled or not (different from hardwaresensor level, will not wait for data)
				
				RobotQuantity quantsJacob; ///< quantities positions in the sensor jacobians
				RobotQuantity quantsState; ///< quantities positions in the filter state

				static IdFactory sensorIds;

			public:
				
				/**
				 * Selectable LOCAL or REMOTE pose constructor.
				 * Creates a sensor with the pose indexed in a map.
				 * \param robPtr the robot
				 * \param poseInFilter flag indicating if the sensor pose is part of the filter (REMOTE).
				 */
				SensorAbstract(const robot_ptr_t & _robPtr, const filtered_obj_t poseInFilter = UNFILTERED, int extraStateFilterSize = 0,
					std::vector<RobotAbstract::Quantity> robQuant = createQuantPosOriQuat());

				/**
				 * Mandatory virtual destructor.
				 */
				virtual ~SensorAbstract() {
				}

				void setIntegrationPolicy(bool integrate_all) { this->integrate_all = integrate_all; }
				bool getIntegrationPolicy() { return integrate_all; }
				void setUseForInit(bool use_for_init) { this->use_for_init = use_for_init; }
				bool getUseForInit() { return use_for_init; }
				void setNeedInit(bool need_init) { this->need_init = need_init; }
				bool getNeedInit() { return need_init; }

				/**
				 * Move function for sensors for now just to apply constraints, but should
				 * be thought more generally it it can have a command, or slowly change (noise), etc
				 * should be included in the state of the robot probably, and anyway done in a more efficient and clean way
				 */
				virtual void move()
				{
					if (poseInFilter)
					{
						jblas::mat QN_q(4,4), XNEW_x(7,7); XNEW_x.clear();
						jblas::sym_mat Q(7,7); Q.clear();

						jblas::vec q = ublas::subrange(pose.x(), 3,7);
						ublasExtra::normalizeJac(q, QN_q);
						ublasExtra::normalize(q);
						ublas::subrange(pose.x(), 3,7) = q;

						ublas::subrange(XNEW_x, 0,3, 0,3) = jblas::identity_mat(3);
						ublas::subrange(XNEW_x, 3,7, 3,7) = QN_q;
						robotPtr()->mapPtr()->filterPtr->predict(robotPtr()->mapPtr()->ia_used_states(), XNEW_x, pose.ia(), Q); // P = F*P*F' + Q
					}
				}

				virtual int queryAvailableRaws(RawInfos &infos) = 0; ///< get information about the available raws and the estimated dates for next one
				virtual int queryNextAvailableRaw(RawInfo &info) = 0; ///< get information about the next available raw
				virtual double getRawTimestamp(unsigned id) = 0;
				virtual void process(unsigned id, double date_limit = -1.) = 0; ///< process the given raw and throw away the previous unprocessed ones \return innovation
				virtual void process_fake(unsigned id, bool move) = 0; ///< don't do any predict or update, but let the data acquisition run smoothly
				virtual void discard(unsigned id) = 0; ///< discard a data without using it
				virtual void init(double date) { use_for_init = false; } ///< use previous data to initialize the robot if needed
				virtual void start() = 0;
				virtual void stop() = 0;
				virtual bool join(int timed_ms = -1) = 0;
				void enable(bool enabled = true) { this->enabled = enabled; if (!enabled) enableHard(false); }
				bool isEnabled() { return enabled; }
				virtual void enableHard(bool enabled = true) = 0;
				virtual void stopDumping() = 0;
				virtual void showInfos() = 0;

				// to replace (because does not work with gcc 4.4) boost::assign::list_of(RobotAbstract::qPos)(RobotAbstract::qOriQuat)
				static std::vector<RobotAbstract::Quantity> createQuantPosOriQuat()
				{
					std::vector<RobotAbstract::Quantity> quant;
					quant.push_back(RobotAbstract::qPos);
					quant.push_back(RobotAbstract::qOriQuat);
					return quant;
				}

				enum type_enum {
					PINHOLE, PINHOLESTEREO, BARRETO
				} type;
				enum kind_enum {
					PROPRIOCEPTIVE, EXTEROCEPTIVE
				} kind;
				virtual std::string categoryName() const {
					return "SENSOR";
				}

				inline void setId() { id(sensorIds.getId()); }
				
				void setPose(double x, double y, double z, double rollDeg,
				    double pitchDeg, double yawDeg);
				void setPoseStd(double x, double y, double z, double rollDeg,
				    double pitchDeg, double yawDeg, double xStd, double yStd,
				    double zStd, double rollDegStd, double pitchDegStd,
				    double yawDegStd);
				void setPoseStd(double x, double y, double z, double rollDeg,
				    double pitchDeg, double yawDeg, double posStd, double oriDegStd) {
					setPoseStd(x, y, z, rollDeg, pitchDeg, yawDeg, posStd, posStd,
					           posStd, oriDegStd, oriDegStd, oriDegStd);
				}


				/// Sensor pose in robot
				Gaussian pose;

				/// Indices of sensor's global pose in map (this is either the \a ia of the robot, \a rob.ia to make it short, or the \b ia_union() of \a rob.ia and \a sen.ia)
				ind_array ia_globalPose;
				/// Flag indicating if the sensor pose is being filtered
				bool poseInFilter;

				size_t missed_data; ///< how many data have been missed before last integration

				/**
				 * Get sensor pose in global frame.
				 * This function composes robot pose with sensor pose to obtain the global sensor pose.
				 *
				 * \return the global pose.
				 */
				vec7 globalPose(bool normalize = false);

				/**
				 * Get sensor pose in global frame.
				 * This function composes robot pose with sensor pose to obtain the global sensor pose.
				 * It renders the Jacobians of the composed frame wrt all variables that are in the map (either robot only, or robot and sensor),
				 * depending on the sensor pose storage being LOCAL or REMOTE (see class Gaussian for doc on storage options).
				 * Therefore, this Jacobian is either 7-by-7 (LOCAL sensor pose) or 7-by-14 (REMOTE sensor pose).
				 *
				 * The concerned states are available as an indirect array \a ia_rs, member of the class and defined at construction time.
				 * \param poseG the global pose.
				 * \param PG_rs the Jacobian wrt the mapped states of robot and sensor.
				 */
				void globalPose(jblas::vec7 & senGlobalPose, jblas::mat & SG_rs);

				virtual void writeLogHeader(kernel::DataLogger& log) const;
				virtual void writeLogData(kernel::DataLogger& log) const;
		};
		
		
		/** 
			Base class for proprioceptive sensors (that directly observe the state of the robot)
			(IMU, odometry, GPS, movement models ...)
			\ingroup rtslam
		*/
		class SensorProprioAbstract: public SensorAbstract
		{
			protected:
				RawVec reading;
			public:
				hardware::hardware_sensorprop_ptr_t hardwareSensorPtr;
			public:
				SensorProprioAbstract(const robot_ptr_t & robPtr, const filtered_obj_t poseInFilter = UNFILTERED, int extraStateFilterSize = 0,
					std::vector<RobotAbstract::Quantity> robQuant = createQuantPosOriQuat()):
				  SensorAbstract(robPtr, poseInFilter, extraStateFilterSize, robQuant) { kind = PROPRIOCEPTIVE; }

				void setHardwareSensor(hardware::hardware_sensorprop_ptr_t hardwareSensorPtr_)
					{ hardwareSensorPtr = hardwareSensorPtr_; }
				virtual void start() { hardwareSensorPtr->start(); }
				virtual void stop() { hardwareSensorPtr->stop(); }
				virtual bool join(int timed_ms = -1) { return hardwareSensorPtr->join(timed_ms); }
				virtual void enableHard(bool enabled = true) { hardwareSensorPtr->enable(enabled); if (enabled) enable(true); }
				virtual void stopDumping() { hardwareSensorPtr->stopDumping(); }
				virtual void showInfos() { hardwareSensorPtr->showInfos(); }


				virtual int queryAvailableRaws(RawInfos &infos)
					{ int res = hardwareSensorPtr->getUnreadRawInfos(infos); infos.integrate_all = integrate_all; return res; }
				virtual int queryNextAvailableRaw(RawInfo &info)
					{ return hardwareSensorPtr->getNextRawInfo(info); }
				virtual double getRawTimestamp(unsigned id) { return hardwareSensorPtr->getRawTimestamp(id); } 
				//process(id) will do the filtering, so it is specific to each hardware
				void process_fake(unsigned id, bool move) { hardwareSensorPtr->getRaw(id, reading); if (move) robotPtr()->move_fake(reading.data(0)); use_for_init = false; }
				void discard(unsigned id) { hardwareSensorPtr->getRaw(id, reading); }
		};
		
		/** 
			Base class for proprioceptive sensors that can be used for filter prediction step.
			They must be able to predict the full pose, and be faster than all other sensors used
			as observations or easily interpolated to simulate it.
			Usually a movement model (eg constant velocity), odometry, 6D IMU...
			\ingroup rtslam
		*/
		class SensorProprioPredictorAbstract: public SensorProprioAbstract
		{
			public:
				SensorProprioPredictorAbstract(const robot_ptr_t & robPtr, const filtered_obj_t poseInFilter = UNFILTERED, int extraStateFilterSize = 0):
					SensorProprioAbstract(robPtr, poseInFilter, extraStateFilterSize) {}
				virtual void predict(double t) = 0;
		
		};
		
		/** 
			Base class for exteroceptive sensors (that need to map the environment)
			(cameras, lasers, ...)
			\ingroup rtslam
		*/
		class SensorExteroAbstract: public SensorAbstract,
		                            public ParentOf<DataManagerAbstract>
		{
			protected:
				// define the type DataManagerList, and the function dataManagerList().
				ENABLE_ACCESS_TO_CHILDREN(DataManagerAbstract,DataManager,dataManager);
			
				raw_ptr_t rawPtr;
			public:
				hardware::hardware_sensorext_ptr_t hardwareSensorPtr;
			
			public:
				
				SensorExteroAbstract(const robot_ptr_t & robPtr, const filtered_obj_t poseInFilter = UNFILTERED, int extraStateFilterSize = 0):
					SensorAbstract(robPtr, poseInFilter, extraStateFilterSize)
				{
					kind = EXTEROCEPTIVE; 
					rawCounter = 0;
				}

				unsigned rawCounter;

				void setHardwareSensor(hardware::hardware_sensorext_ptr_t hardwareSensorPtr_)
					{ hardwareSensorPtr = hardwareSensorPtr_; }
				virtual void start() { hardwareSensorPtr->start(); }
				virtual void stop() { hardwareSensorPtr->stop(); }
				virtual bool join(int timed_ms = -1) { return hardwareSensorPtr->join(timed_ms); }
				virtual void enableHard(bool enabled = true) { hardwareSensorPtr->enable(enabled); if (enabled) enable(true); }
				virtual void stopDumping() { hardwareSensorPtr->stopDumping(); }
				virtual void showInfos() { hardwareSensorPtr->showInfos(); }

//				virtual int acquireRaw() = 0;
//				virtual raw_ptr_t getRaw() = 0;

				virtual raw_ptr_t getLastProcessedRaw() { raw_ptr_t raw; hardwareSensorPtr->getLastProcessedRaw(raw); return raw; }
				virtual int queryAvailableRaws(RawInfos &infos)
					{ int res = hardwareSensorPtr->getUnreadRawInfos(infos); infos.integrate_all = integrate_all; return res; }
				virtual int queryNextAvailableRaw(RawInfo &info)
					{ return hardwareSensorPtr->getNextRawInfo(info); }
				virtual double getRawTimestamp(unsigned id) { return hardwareSensorPtr->getRawTimestamp(id); } 
				void process(unsigned id, double date_limit = -1.);
				void process_fake(unsigned id, bool move) { hardwareSensorPtr->getRaw(id, rawPtr); if (move) robotPtr()->move_fake(rawPtr->timestamp); rawCounter++; }
				void discard(unsigned id) { hardwareSensorPtr->getRaw(id, rawPtr); }
		};

	}
}

#endif /* SENSORABSTRACT_H_ */
