/**
 * \file HardwareSensorCamera.cpp
 * \date 18/06/2010
 * \author croussil
 * \ingroup rtslam
 */

#include <algorithm>
#include <sstream>
#include <fstream>

#if 1
// creates conflict with boost sandbox with boost 1.42 in debug
//#include <boost/bind.hpp>
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>
#endif

#include "kernel/timingTools.hpp"
#include "rtslam/hardwareSensorCamera.hpp"


#include <image/Image.hpp>

namespace jafar {
namespace rtslam {
namespace hardware {

  namespace bfs = boost::filesystem;

  #if BOOST_VERSION > 104600
  #define STRING_GETTER .string()
  #else
  #define STRING_GETTER
  #endif

	void HardwareSensorCamera::preloadTaskOffline(void)
	{ JFR_GLOBAL_TRY
		unsigned findex;
		dump_path = dump_path + "/"; // TODO check if necessary

		// find image files ; maybe for very long sequences it would be better to just find first and last, to avoid storing the whole file list
		std::string prefix;
		std::string file_dates;
		std::set<bfs::path> file_names;
		for(int i = 0; i < 2; ++i)
		{
			std::ostringstream ossp; ossp << "image";
			if (i == 0) ossp << "_" << id();
			file_dates = ossp.str() + ".dat";
			ossp << "_";
			prefix = ossp.str();
			ossp << "([0-9]*)\\.(pgm|ppm|png|bmp|tif|tiff|jpg)";
			boost::regex pattern(ossp.str());

			for (bfs::directory_iterator iter(bfs::path(dump_path.c_str())), end; iter != end; ++iter)
			{
				std::string name = iter->path().filename()STRING_GETTER;
				if (boost::regex_match(name, pattern))
				{
					if (atoi(boost::regex_replace(name, pattern, "$1").c_str()) % filter_div == filter_mod)
						file_names.insert(iter->path().filename());//dump_path + std::string("/") + name);
				}
			}

			if (file_names.size() != 0 || id() != 1) break;
		}
		if (file_names.size() == 0)
			std::cout << "Did not find any image" << std::endl; else
			std::cout << "First image " << file_names.begin()->string() << " ; Last image " << file_names.rbegin()->string() << std::endl;

		// open binary dates file
		std::fstream fb(file_dates.c_str(), std::ios_base::in | std::ios_base::binary);
		bool binary = !fb.fail();

		// now iterate on images
		std::set<bfs::path>::iterator iter;
		for (iter = file_names.begin(); iter != file_names.end() && !stopping; ++iter)
		{
			// acquire the image
			boost::unique_lock<boost::mutex> l(mutex_data);
			while (isFull(true) && !stopping) cond_offline_freed.wait(l);
			l.unlock();
			if (stopping) break;
			int buff_write = getWritePos();
			while (!stopping)
			{
				// load image
				if (!bufferSpecPtr[buff_write]->img->load(dump_path + iter->string(), 0) || bufferSpecPtr[buff_write]->img->data() == NULL) continue;
				index = atoi(iter->string().substr(prefix.length(), iter->stem()STRING_GETTER.length()-prefix.length()).c_str());

				// load date
				bufferSpecPtr[buff_write]->name(iter->string());
				if (binary)
				{
					do {
						fb.read((char*)&findex, sizeof(unsigned));
						fb.read((char*)&(bufferSpecPtr[buff_write]->timestamp), sizeof(double));
					} while (findex < index && !fb.eof());
					if (!fb && fb.gcount() < (int)sizeof(double))
						JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "HardwareSensorCamera::preloadTaskOffline : binary timestamps file truncated.");
				}
				else
				{
					std::fstream f((dump_path + (*iter).stem()STRING_GETTER+std::string(".time")).c_str(), std::ios_base::in);
					f >> bufferSpecPtr[buff_write]->timestamp; f.close();
				}
				break;
			}
			incWritePos();
			if (condition) condition->setAndNotify(1);
		}
		if (iter == file_names.end())
		{
			boost::unique_lock<boost::mutex> l(mutex_data);
			no_more_data = true;
			//std::cout << "No more images to read." << std::endl;
		}

		if (binary) fb.close();

		JFR_GLOBAL_CATCH
	}


	void HardwareSensorCamera::init(Mode mode, std::string dump_path, cv::Size imgSize)
	{
		this->mode = mode;
		this->dump_path = dump_path;

		// configure data
		bufferImage.resize(bufferSize);
		bufferSpecPtr.resize(bufferSize);
		for(int i = 0; i < bufferSize; ++i)
		{
			bufferImage[i] = cvCreateImage(imgSize, 8, 1);
			buffer(i).reset(new RawImage());
			bufferSpecPtr[i] = SPTR_CAST<RawImage>(buffer(i));
			bufferSpecPtr[i]->setJafarImage(jafarImage_ptr_t(new image::Image(bufferImage[i])));
		}
		
		found_first = 0;
		index = 0;
	}


	void HardwareSensorCamera::start()
	{
		if (started) { std::cout << "Warning: HardwareSensorCamera " << id() << " has already been started" << std::endl; return; }

		// start acquire task
		last_timestamp = kernel::Clock::getTime();
		if (mode == mOffline)
			preloadTask_thread = new boost::thread(boost::bind(&HardwareSensorCamera::preloadTaskOffline,this));
		else
			preloadTask_thread = new boost::thread(boost::bind(&HardwareSensorCamera::preloadTask,this));

		started = true;
	}

	void HardwareSensorCamera::stop()
	{
		if (!started) return;
		stopping = true;
		cond_offline_freed.notify_all();
	}

	bool HardwareSensorCamera::join(int timed_ms)
	{
		if (boost::this_thread::get_id() != preloadTask_thread->get_id())
		{
			if (timed_ms < 0)
				preloadTask_thread->join(); else
				return preloadTask_thread->timed_join(boost::posix_time::milliseconds(timed_ms));
		}
		return true;
	}

	HardwareSensorCamera::HardwareSensorCamera(kernel::VariableCondition<int> *condition, int cam_id, cv::Size imgSize, std::string dump_path):
		HardwareSensorExteroAbstract(condition, mOffline, 3), cam_id(cam_id), filter_div(1), filter_mod(0)
	{
		init(mOffline, dump_path, imgSize);
	}

	HardwareSensorCamera::HardwareSensorCamera(kernel::VariableCondition<int> *condition, Mode mode, int cam_id, int bufferSize, kernel::LoggerTask *loggerTask):
		HardwareSensorExteroAbstract(condition, mode, bufferSize), loggerTask(loggerTask), cam_id(cam_id), filter_div(1), filter_mod(0)
	{}

	

}}}

