/**
 * \file HardwareSensorCameraUeye.cpp
 * \date 18/06/2010
 * \author croussil
 * \ingroup rtslam
 */

#include <algorithm>
#include <sstream>
#include <fstream>

#if 0
// creates conflict with boost sandbox with boost 1.42 in debug
#include <boost/bind.hpp>
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>
#endif

#include "kernel/timingTools.hpp"
#include "rtslam/hardwareSensorCameraUeye.hpp"


#include <image/Image.hpp>

#ifdef HAVE_UEYE
/// convert UEYETIME to double epoch
double convertUeyeTime(UEYETIME &ueyeTime)
{
	struct tm t;
	t.tm_year = ueyeTime.wYear-1900;
	t.tm_mon = ueyeTime.wMonth-1;
	t.tm_mday = ueyeTime.wDay;
	t.tm_hour = ueyeTime.wHour;
	t.tm_min = ueyeTime.wMinute;
	t.tm_sec = ueyeTime.wSecond;
	t.tm_isdst = -1;
	return mktime(&t) + ueyeTime.wMilliseconds * 1e-3;
}
#endif

namespace jafar {
namespace rtslam {
namespace hardware {

	void HardwareSensorCameraUeye::preloadTask(void)
	{ JFR_GLOBAL_TRY
#ifdef HAVE_UEYE
		char *image;
		int imageID;
		UEYEIMAGEINFO imageInfo;
#endif
		index = 0;

		std::fstream fb;
		if (mode == mOnlineDump || mode == mOffline)
		{
			std::ostringstream oss; oss << dump_path << "/images";
			if (id()) oss << "_" << id();
			oss << ".dat";
			fb.open(oss.str().c_str(), std::ios_base::out | std::ios_base::binary);
		}

		while(!stopping)
		{
			// acquire the image
#ifdef HAVE_UEYE
			int buff_write = getWritePos(true); // don't need to lock because we are the only writer

			if (is_WaitForNextImage(camera, 1000, &image, &imageID) != IS_SUCCESS) continue;
			bufferSpecPtr[buff_write]->arrival = kernel::Clock::getTime();
			// TODO would be better if avoiding copy here
			memcpy(bufferImage[buff_write]->imageData, image, bufferImage[buff_write]->imageSize);
			if (is_UnlockSeqBuf(camera, imageID, image) != IS_SUCCESS)  { std::cerr << "HardwareSensorCameraUeye: unlock failed" << std::endl; }

			if (is_GetImageInfo(camera, imageID, &imageInfo, sizeof(imageInfo)) != IS_SUCCESS) continue;
			bufferSpecPtr[buff_write]->timestamp = convertUeyeTime(imageInfo.TimestampSystem);
			if (enabled)
			{
				update_arrival_delay(bufferSpecPtr[buff_write]->arrival - bufferSpecPtr[buff_write]->timestamp);
				// TODO use camera timestamp to filter pc timestamp
				//double cameraTimeStamp = imageInfo.u64TimestampDevice * 1e-7;
				//printf("%.16g\t%.19g\t%.12g\n", bufferSpecPtr[buff_write]->timestamp, bufferSpecPtr[buff_write]->arrival, imageInfo.u64TimestampDevice * 1e-7);
				last_timestamp = bufferSpecPtr[buff_write]->timestamp;
				incWritePos();
				if (condition) condition->setAndNotify(1);
			}

			if (mode == mOnlineDump)
			{
				rawimage_ptr_t img = rawimage_ptr_t(static_cast<RawImage*>(bufferSpecPtr[buff_write]->clone()));
				loggerTask->push(new LoggableImage(dump_path, img, index, (id() ? id() : -1), fb));
			}
			++index;
#endif
		}

		if (mode == mOnlineDump) { loggerTask->push(new LoggableClose(fb)); while (fb.is_open()) usleep(100000); }
		if (mode == mOffline) fb.close();

		JFR_GLOBAL_CATCH
	}

	
	HardwareSensorCameraUeye::HardwareSensorCameraUeye(kernel::VariableCondition<int> *condition, int cam_id, cv::Size imgSize, std::string dump_path):
		HardwareSensorCamera(condition, cam_id, imgSize, dump_path)
	{}
	

#ifdef HAVE_UEYE
	void HardwareSensorCameraUeye::init(const std::string &camera_id, cv::Size imgSize, double shutter, double freq, int trigger, Mode mode, std::string dump_path)
	{
		int r;
		char *msg;
		// configure camera
		if (mode == mOnline || mode == mOnlineDump)
		{
			// configure
			int ncameras;
			if (is_GetNumberOfCameras(&ncameras) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_GetNumberOfCameras: " << msg << std::endl; return; }
			if (ncameras <= 0) { std::cerr << "HardwareSensorCameraUeye::init: Could not find camera" << std::endl; return; }
			PUEYE_CAMERA_LIST camList = (PUEYE_CAMERA_LIST) new char[sizeof(DWORD) + ncameras * sizeof(UEYE_CAMERA_INFO)];
			camList->dwCount = ncameras;
			if (is_GetCameraList(camList) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_GetCameraList: " << msg << std::endl; return; }
			camera = 0; // use first camera
			if (is_InitCamera(&camera, NULL) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_InitCamera: " << msg << std::endl; return; }
			SENSORINFO sensorInfo;
			if (is_GetSensorInfo(camera, &sensorInfo) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_GetSensorInfo: " << msg << std::endl; return; }
			std::cout << "HardwareSensorCameraUeye: Using sensor " << sensorInfo.strSensorName << " ID " << camList->uci[camera].SerNo << std::endl;
			if (sensorInfo.bGlobShutter == FALSE) { std::cerr << "HardwareSensorCameraUeye: Cannot use a rolling shutter camera" << std::endl; return; }
			IS_RECT rect;
			rect.s32X = (sensorInfo.nMaxWidth-imgSize.width)/2; rect.s32Width = imgSize.width;
			rect.s32Y = (sensorInfo.nMaxHeight-imgSize.height)/2; rect.s32Height = imgSize.height;
			if (is_AOI(camera, IS_AOI_IMAGE_SET_AOI, (void*)&rect, sizeof(IS_RECT)) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_AOI(IS_AOI_IMAGE_SET_AOI): " << msg << std::endl; return; }
			if (is_SetColorMode(camera, IS_CM_MONO8 ) != IS_SUCCESS )
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_SetColorMode(IS_CM_MONO8): " << msg << std::endl; return; }
			if (is_SetFrameRate(camera, freq, &realFreq) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_SetFrameRate: " << msg << std::endl; return; }
			if (trigger != 0) realFreq = freq;
			std::cout << "Camera set to freq " << realFreq << " Hz (external trigger " << trigger << ")" << std::endl;
			if (shutter > 1e-6)
			{
				shutter *= 1000; // convert to ms
				if (is_SetAutoParameter(camera, IS_SET_AUTO_SHUTTER_MAX, &shutter, NULL) != IS_SUCCESS) // must be before ENABLE_AUTO_SHUTTER
					{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_SetAutoParameter(IS_SET_AUTO_SHUTTER_MAX): " << msg << std::endl; return; }
			}
			double exposure = 1;
			if (is_SetAutoParameter(camera, IS_SET_ENABLE_AUTO_SHUTTER, &exposure, NULL) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_SetAutoParameter(IS_SET_ENABLE_AUTO_SHUTTER): " << msg << std::endl; return; }
			if (is_SetAutoParameter(camera, IS_SET_ENABLE_AUTO_GAIN, &exposure, NULL) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_SetAutoParameter(IS_SET_ENABLE_AUTO_GAIN): " << msg << std::endl; return; }
			// TODO set trigger
			if (is_SetExternalTrigger(camera, IS_SET_TRIGGER_OFF) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_SetExternalTrigger: " << msg << std::endl; return; }

			for (int i = 0; i < 2; ++i)
			{
				char* imageMem;
				int pid;

				if (is_AllocImageMem(camera, imgSize.width, imgSize.height, 8, &imageMem, &pid) != IS_SUCCESS)
					{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_AllocImageMem: " << msg << std::endl; return; }
				if( is_AddToSequence(camera, imageMem, pid) != IS_SUCCESS)
					{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_AddToSequence: " << msg << std::endl; return; }
			}

			//int widthStep;
			//if (is_GetImageMemPitch(camera, &widthStep) != IS_SUCCESS) { std::cerr << "HardwareSensorCameraUeye: Could not get widthStep" << std::endl; return; }
			//if (widthStep != bufferImage[0]->widthStep) { std::cerr << "HardwareSensorCameraUeye: uEye widthStep " << widthStep << " != opencv widthStep " << bufferImage[0]->widthStep << std::endl; return; }
			if (is_InitImageQueue (camera, 0) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_InitImageQueue: " << msg << std::endl; return; }
			if (is_CaptureVideo(camera, IS_DONT_WAIT) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::init, is_CaptureVideo: " << msg << std::endl; return; }
		}

		HardwareSensorCamera::init(mode, dump_path, imgSize);
	}

	HardwareSensorCameraUeye::HardwareSensorCameraUeye(kernel::VariableCondition<int> *condition, int bufferSize, const std::string &camera_id,
		cv::Size size, double freq, int trigger, double shutter, Mode mode, int cam_id, std::string dump_path, kernel::LoggerTask *loggerTask):
		HardwareSensorCamera(condition, mode, cam_id, bufferSize, loggerTask)
	{
		if (mode == mOnlineDump && !loggerTask) JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "HardwareSensorCameraUeye: you must provide a loggerTask if you want to dump data.");
		init(camera_id, size, shutter, freq, trigger, mode, dump_path);
	}
#endif


	HardwareSensorCameraUeye::~HardwareSensorCameraUeye()
	{
#ifdef HAVE_UEYE
		int r;
		char *msg;
		if (mode == mOnline || mode == mOnlineDump)
		{
			if (is_StopLiveVideo(camera, IS_WAIT) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::~, is_StopLiveVideo: " << msg << std::endl; }
			if (is_ExitImageQueue(camera) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::~, is_ExitImageQueue: " << msg << std::endl; }
			if (is_ClearSequence(camera) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::~, is_ClearSequence: " << msg << std::endl; }
			// TODO get back buffers to release them, or store them at creation
			//for (int i = 0; i < 2; ++i) is_FreeImageMem(camera, , );
			if (is_ExitCamera(camera) != IS_SUCCESS)
				{ is_GetError (camera, &r, &msg); std::cerr << "HardwareSensorCameraUeye::~, is_ExitCamera: " << msg << std::endl; }
		}
#endif
	}


}}}

