/**
 * \file hardwareSensorGpsGenom.cpp
 *
 * File for getting data from the genom gps module
 *
 * \date 16/03/2011
 * \author croussil
 *
 * \ingroup rtslam
 */

#include "kernel/timingTools.hpp"
#include "rtslam/hardwareSensorGpsGenom.hpp"

#ifdef HAVE_POSTERLIB
#include "h2timeLib.h"
#endif




namespace jafar {
namespace rtslam {
namespace hardware {


	void HardwareSensorGpsGenom::preloadTask(void)
	{ JFR_GLOBAL_TRY
		char data[256];
#ifdef HAVE_POSTERLIB
		H2TIME h2timestamp;
#endif
		unsigned long prev_ntick = 0;
		double *date, prev_date = 0.0;
		double *pos; float *var;
		int format, size;
		
		if (mode == mOnlineDump) log.openWrite("GPS", dump_path, 0, reading.data.size(), loggerTask);
		if (mode == mOffline) log.openRead("GPS", dump_path, format, size);
		
		while (!stopping)
		{
			if (mode == mOffline)
			{
				bool eof = log.read(reading.data);
				boost::unique_lock<boost::mutex> l(mutex_data);
				if (isFull(true)) cond_offline_full.notify_all();
				if (eof) { no_more_data = true; cond_offline_full.notify_all(); stopping = true; }
				while (isFull(true) && !stopping) cond_offline_freed.wait(l);
				if (stopping) break;
			} else
			{
#ifdef HAVE_POSTERLIB
				while (!stopping) // wait for new data
				{
					usleep(2000);
					if (reentrantPosterIoctl(posterId, FIO_GETDATE, &h2timestamp) != ERROR)
					{
						if (h2timestamp.ntick != prev_ntick)
						{
							prev_ntick = h2timestamp.ntick;
							if (reentrantPosterRead(posterId, 0, data, 256) != ERROR)
							{
								date = (double*)(data+80+48);
								if (*date != prev_date)
								{
									prev_date = *date;
									break;
								}
							}
						}
					}
				}
#endif
				reading.arrival = kernel::Clock::getTime();
				pos = (double*)(data+16);
				var = (float*)(data+48);
				reading.data(0) = *date;
				reading.data(1) = pos[1]; // swap
				reading.data(2) = pos[0];
				reading.data(3) = pos[2];
				reading.data(4) = var[1]; // swap
				reading.data(5) = var[0];
				reading.data(6) = var[2];
				if (*(unsigned char*)(data+73) < 5) continue; // because there is no redundancy with only 4 sats -> risk of inconsistency with reflections

				if (isFull())
				{
					std::cout << std::setprecision(16) << reading.arrival << ": data not read, will clear the buffer to store new data at " << std::setprecision(16) << reading.data(0) << " (increase GPS buffer size)" << std::endl;
					releaseUntil(getLastUnreadPos());
					//JFR_ERROR(RtslamException, RtslamException::BUFFER_OVERFLOW, "Data not read: Increase GPS buffer size !");
				}

				if (reading.arrival-reading.data(0) > 3) continue; // discard data too old

				//std::cout << "GPS poster : " << std::setprecision(15) << reading.data << std::endl;
			}
			
			if (enabled)
			{
				int buff_write = getWritePos(true); // don't need to lock because we are the only writer
				buffer(buff_write) = reading;
				buffer(buff_write).data(0) += timestamps_correction;
				update_arrival_delay(reading.arrival - buffer(buff_write).data(0));
				last_timestamp = buffer(buff_write).data(0);
				incWritePos();
				if (condition) condition->setAndNotify(1);
			}

			if (mode == mOnlineDump) log.write(reading.data);
		}

		if (mode == mOnlineDump || mode == mOffline) log.close();

		JFR_GLOBAL_CATCH
	}
	
	
	HardwareSensorGpsGenom::HardwareSensorGpsGenom(kernel::VariableCondition<int> *condition, unsigned bufferSize, Mode mode, std::string dump_path, kernel::LoggerTask *loggerTask):
		HardwareSensorProprioAbstract(condition, mode, bufferSize, ctVar), dump_path(dump_path), loggerTask(loggerTask)
	{
		if (mode == mOnlineDump && !loggerTask) JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "HardwareSensorGpsGenom: you must provide a loggerTask if you want to dump data.");
		addQuantity(qPos);
		//addQuantity(qAbsVel);
		initData();

		// configure
		if (mode == mOnline || mode == mOnlineDump)
		{
#ifdef HAVE_POSTERLIB
			if (posterFind("GPSInfo", &this->posterId) == ERROR) {
				JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "Poster GPSInfo could not be found");
			}
#else
			JFR_ERROR(RtslamException, RtslamException::MISSING_DEPENDENCY,
			          "You must install posterlib (pocolibs) if you want to use online GPS sensor");
#endif
		}
	}

	void HardwareSensorGpsGenom::start()
	{
		// start acquire task
		//preloadTask();
		if (started) { std::cout << "Warning: This HardwareSensorGpsGenom has already been started" << std::endl; return; }
		started = true;
		last_timestamp = kernel::Clock::getTime();
		preloadTask_thread = new boost::thread(boost::bind(&HardwareSensorGpsGenom::preloadTask,this));
		
		if (mode == mOffline)
		{ // wait that log has been read before first frame
			boost::unique_lock<boost::mutex> l(mutex_data);
			cond_offline_full.wait(l);
		}
	}

	void HardwareSensorGpsGenom::stop()
	{
		if (!started) return;
		stopping = true;
		cond_offline_freed.notify_all();
	}

	bool HardwareSensorGpsGenom::join(int timed_ms)
	{
		if (boost::this_thread::get_id() != preloadTask_thread->get_id())
		{
			if (timed_ms < 0)
				preloadTask_thread->join(); else
				return preloadTask_thread->timed_join(boost::posix_time::milliseconds(timed_ms));
		}
		return true;
	}


}}}

