/**
 * \file hardwareSensorOdomRmp400Genom.cpp
 *
 * File for getting odometry data from the rmp400 module
 *
 * \date 16/03/2011
 * \author croussil
 *
 * \ingroup rtslam
 */

#include "kernel/timingTools.hpp"
#include "rtslam/hardwareSensorOdomRmp400Genom.hpp"

#ifdef HAVE_POSTERLIB
#include "h2timeLib.h"
#endif


namespace jafar {
namespace rtslam {
namespace hardware {

/*******************************/
/* INCLUDE FROM rmp400Struct.h */

// WARNING Alignment is different on 32bits and 64bits !

typedef enum RMP400_MODE {
	RMP400_IDLE = 0,
	RMP400_MOTORS_OFF,
	RMP400_MANUAL,
	RMP400_TRACK,
	RMP400_EMERGENCY
} RMP400_MODE;

typedef struct RMP400_DATA_STR {
	double pitch_angle;
	double pitch_rate;
	double roll_angle;
	double roll_rate;
	double lw_velocity;
	double rw_velocity;
	double yaw_rate;
	short servo_frames;
	double integrated_left_wheel;
	double integrated_right_wheel;
	double integrated_fore_aft;
	double integrated_yaw;
	double left_torque;
	double right_torque;
	short operational_mode;
	short controller_gain_schedule;
	double ui_voltage;
	double powerbase_voltage;
	double battery_charge;
	double velocity_command;
	double turn_command;
} RMP400_DATA_STR;

typedef struct RMP400_STATUS_STR {
	RMP400_MODE rs_mode;
	RMP400_DATA_STR rs_data[2];
} RMP400_STATUS_STR;

// new generic version for RMP400 and RMP440

typedef enum RMP440_MODE {
	RMP440_IDLE = 0,
	RMP440_MOTORS_OFF,
	RMP440_MANUAL,
	RMP440_TRACK,
	RMP440_EMERGENCY,
	RMP440_POWER_CORD
} RMP440_MODE;

typedef enum RMP_MODEL {
	RMP_MODEL_400 = 0,
	RMP_MODEL_440
} RMP_MODEL;

typedef struct RMP_STATUS_STR {
	double receive_date;
	RMP_MODEL rmp_model;
	RMP440_MODE rs_mode;
	double propulsion_battery_level;
	double aux_battery_level;
	double pitch, roll, yaw_rate;
	double v, w;
	double v_target, w_target;
	float right_front_vel, left_front_vel, right_rear_vel, left_rear_vel;
	float right_front_pos, left_front_pos, right_rear_pos, left_rear_pos;
	float right_front_torque, left_front_torque, right_rear_torque, left_rear_torque;
} RMP_STATUS_STR;


/*******************************/

	enum ReadingData { rdTime=0, rdVel, rdUncertainty, rdVelLF, rdVelRF, rdVelLB, rdVelRB, rdTorqLF, rdTorqRF, rdTorqLB, rdTorqRB, rdVelCmd, rdTurnCmd, NReadingData };

	void HardwareSensorOdomRmp400Genom::filterOdom(jblas::vec & reading)
	{		
		// compute velocity norm
		double vel = fabs(reading(rdVelLF) + reading(rdVelRF) + reading(rdVelLB) + reading(rdVelRB)) / 4.;

		// additive and multiplicative bias
		double uncert = 0.01;
		uncert += vel * 0.005;

		// penalty for heterogeneous wheels velocitie
		double lw_diff = fabs(reading(rdVelLF)-reading(rdVelLB));
		double rw_diff = fabs(reading(rdVelRF)-reading(rdVelRB));
		uncert += lw_diff;
		uncert += rw_diff;
		if (vel > 1e-2 && (lw_diff > vel || rw_diff > vel)) uncert = 1e10;

		// penalty for high torques
		// TODO

		// penalty for heterogeneous torques
		// TODO

		// penalty for turn radius
		//double turn = reading(rdTurnCmd); // broken
		double turn = fabs((reading(rdVelLF)+reading(rdVelLB))-(reading(rdVelRF)+reading(rdVelRB))) / 2.0 / 1.0; // should divide by axisWidth instead of 1.0...
		if (turn > 1e-4)
		{
			double rad = vel / turn;
			uncert += 0.1 * exp(-rad);
		}
#if 0
		// penalty for acceleration
		// we could also check rdVelCmd to anticipate
		// and we should smooth that acceleration measure...
		const double accel_ref = 0.5;
		if (prev_reading.size() == 0) uncert = 2e10; else
		{
			double dt = reading(rdTime) - prev_reading(rdTime);
			double max_accel = fabs(vel - prev_reading(rdVel))/dt;
/*			double max_accel = 0.0;
			for(int i = rdVelLF; i < rdVelLF+4; ++i)
			{
				double accel = fabs(reading(i) - prev_reading(rdVelLF))/dt;
				if (accel > max_accel) max_accel = accel;
			}
*/			uncert *= exp(max_accel / accel_ref);
		}
#endif
		// special case when stopped for more than a few seconds
		// chosen these thresholds because of some spurious values of 0.003 when stopped
		// TODO also check that command is 0 and pitch/roll not too high ?
		if (vel < 1e-6)
		{
			if (time_stopped >= 0)
			{
				if (reading(rdTime)-time_stopped >= 2.0)
					uncert = 0.0;
			} else time_stopped = reading(rdTime);
		} else
		{
			if (vel > 4e-3)
				time_stopped = -1;
		 }

		// return results
		// TODO we could also try to correct velocity with the detected faults
		reading(rdVel) = vel;
		reading(rdUncertainty) = uncert;
		prev_reading = reading;
	}


	void HardwareSensorOdomRmp400Genom::preloadTask(void)
	{ JFR_GLOBAL_TRY
		union { RMP400_STATUS_STR data400; RMP_STATUS_STR data; };
#ifdef HAVE_POSTERLIB
		H2TIME h2timestamp;
#endif
		unsigned long prev_ntick = 0;
		double prev_date = 0.0;
		int format, size;

		if (mode == mOnlineDump) log.openWrite("odom", dump_path, 0, reading.data.size(), loggerTask);
		if (mode == mOffline) log.openRead("odom", dump_path, format, size);

		while (!stopping)
		{
			if (mode == mOffline)
			{
				bool eof = log.read(reading.data);
				boost::unique_lock<boost::mutex> l(mutex_data);
				if (isFull(true)) cond_offline_full.notify_all();
				if (eof) { no_more_data = true; cond_offline_full.notify_all(); stopping = true; }
				while (isFull(true) && !stopping) cond_offline_freed.wait(l);
				if (stopping) break;
			} else
			{
#ifdef HAVE_POSTERLIB
				while (!stopping) // wait for new data
				{
					usleep(2000);
					if (reentrantPosterIoctl(posterId, FIO_GETDATE, &h2timestamp) != ERROR)
					{
						if (h2timestamp.ntick != prev_ntick)
						{
							prev_ntick = h2timestamp.ntick;

							if (status400poster && reentrantPosterRead(posterId, 0, &data400, sizeof(RMP400_STATUS_STR)) != ERROR)
							{
								double val = data400.rs_data[0].pitch_rate + data400.rs_data[1].roll_rate;
								if (prev_date != val) { prev_date = val; break; }
							}
							if (!status400poster && reentrantPosterRead(posterId, 0, &data, sizeof(RMP_STATUS_STR)) != ERROR)
							{
								if (prev_date != data.receive_date) { prev_date = data.receive_date; break; }
							}
						}
					}
				}
#endif
				reading.arrival = kernel::Clock::getTime();
				reading.data[rdVel] = 0.;
				reading.data[rdUncertainty] = 0.;

				if (status400poster)
				{
					reading.data[rdTime] = reading.arrival;
					reading.data[rdVelLF] = data400.rs_data[0].lw_velocity;
					reading.data[rdVelRF] = data400.rs_data[0].rw_velocity;
					reading.data[rdVelLB] = data400.rs_data[1].lw_velocity;
					reading.data[rdVelRB] = data400.rs_data[1].rw_velocity;
					reading.data[rdTorqLF] = data400.rs_data[0].left_torque;
					reading.data[rdTorqRF] = data400.rs_data[0].right_torque;
					reading.data[rdTorqLB] = data400.rs_data[1].left_torque;
					reading.data[rdTorqRB] = data400.rs_data[1].right_torque;
					reading.data[rdVelCmd] = (data400.rs_data[0].velocity_command+data400.rs_data[1].velocity_command)/2.;
					reading.data[rdTurnCmd] = (data400.rs_data[0].turn_command+data400.rs_data[1].turn_command)/2.;
				} else
				{
					reading.data[rdTime] = data.receive_date;
					reading.data[rdVelLF] = data.left_front_vel;
					reading.data[rdVelRF] = data.right_front_vel;
					reading.data[rdVelLB] = data.left_rear_vel;
					reading.data[rdVelRB] = data.right_rear_vel;
					reading.data[rdTorqLF] = data.left_front_torque;
					reading.data[rdTorqRF] = data.right_front_torque;
					reading.data[rdTorqLB] = data.left_rear_torque;
					reading.data[rdTorqRB] = data.right_rear_torque;
					reading.data[rdVelCmd] = data.v_target;
					reading.data[rdTurnCmd] = data.w_target;
				}

				if (isFull())
				{
					std::cout << std::setprecision(16) << reading.arrival << ": data not read, will clear the buffer to store new data at " << std::setprecision(16) << reading.data(0) << " (increase OdomRmp400 buffer size)" << std::endl;
					releaseUntil(getLastUnreadPos());
					//JFR_ERROR(RtslamException, RtslamException::BUFFER_OVERFLOW, "Data not read: Increase GPS buffer size !");
				}

				if (reading.arrival-reading.data[rdTime] > 3) continue; // discard data too old

			}

			reading.data[rdVelLF] *= calib[0];
			reading.data[rdVelRF] *= calib[1];
			reading.data[rdVelLB] *= calib[2];
			reading.data[rdVelRB] *= calib[3];

			filterOdom(reading.data);

			if (enabled)
			{
				int buff_write = getWritePos();
				buffer(buff_write) = reading;
				buffer(buff_write).data(0) += timestamps_correction;
				update_arrival_delay(reading.arrival - buffer(buff_write).data(0));
				last_timestamp = buffer(buff_write).data(0);
				incWritePos();
				if (condition) condition->setAndNotify(1);
			}

			if (mode == mOnlineDump) log.write(reading.data);
		}

		if (mode == mOnlineDump || mode == mOffline) log.close();

		JFR_GLOBAL_CATCH
	}
	
	
	HardwareSensorOdomRmp400Genom::HardwareSensorOdomRmp400Genom(kernel::VariableCondition<int> *condition, unsigned bufferSize,
		Mode mode, std::string dump_path, kernel::LoggerTask *loggerTask):
		HardwareSensorProprioAbstract(condition, mode, bufferSize, ctNone), dump_path(dump_path), loggerTask(loggerTask), time_stopped(-1)
	{
		calib = jblas::scalar_vec(4, 1.0);
		if (mode == mOnlineDump && !loggerTask) JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "HardwareSensorOdomRmp400Genom: you must provide a loggerTask if you want to dump data.");
		addQuantity(qNormVel);
		addDataQuantities(11);
		initData();

		// configure
		if (mode == mOnline || mode == mOnlineDump)
		{
#ifdef HAVE_POSTERLIB
			if (posterFind("rmp400StatusGeneric", &this->posterId) == ERROR &&
					posterFind("rmp440StatusGeneric", &this->posterId) == ERROR) {
				if (posterFind("rmp400Status", &this->posterId) == ERROR)
				{
					JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "Poster rmp400StatusGeneric or rmp440StatusGeneric or rmp400Status could not be found");
				}
				status400poster = true;
			} else status400poster = false;
#else
			JFR_ERROR(RtslamException, RtslamException::MISSING_DEPENDENCY,
								"You must install posterlib (pocolibs) if you want to use online rmp400 odometry sensor");
#endif
		}
	}

	void HardwareSensorOdomRmp400Genom::start()
	{
		// start acquire task
		//preloadTask();
		if (started) { std::cout << "Warning: This HardwareSensorOdomRmp400Genom has already been started" << std::endl; return; }
		started = true;
		last_timestamp = kernel::Clock::getTime();
		preloadTask_thread = new boost::thread(boost::bind(&HardwareSensorOdomRmp400Genom::preloadTask,this));
		
		if (mode == mOffline)
		{ // wait that log has been read before first frame
			boost::unique_lock<boost::mutex> l(mutex_data);
			cond_offline_full.wait(l);
		}
	}

	void HardwareSensorOdomRmp400Genom::stop()
	{
		if (!started) return;
		stopping = true;
		cond_offline_freed.notify_all();
	}

	bool HardwareSensorOdomRmp400Genom::join(int timed_ms)
	{
		if (boost::this_thread::get_id() != preloadTask_thread->get_id())
		{
			if (timed_ms < 0)
				preloadTask_thread->join(); else
				return preloadTask_thread->timed_join(boost::posix_time::milliseconds(timed_ms));
		}
		return true;
	}

	
}}}
