/**
 * \file observationAbstract.cpp
 * \date 10/03/2010
 * \author jsola
 * \ingroup rtslam
 */

#include <boost/numeric/bindings/ublas/matrix.hpp>
#include <boost/numeric/bindings/ublas/vector.hpp>
#include <boost/numeric/bindings/lapack/driver/syev.hpp>
#include <boost/numeric/bindings/ublas/symmetric.hpp>

#include "kernel/jafarDebug.hpp"

#include "rtslam/observationAbstract.hpp"
#include "rtslam/sensorAbstract.hpp"
#include "rtslam/landmarkAbstract.hpp"

#include "rtslam/featureAbstract.hpp"


namespace jafar {
	namespace rtslam {
		using namespace std;


		//////////////////////////
		// OBSERVATION ABSTRACT
		//////////////////////////
		const unsigned ObservationAbstract::max_hypotheses = 10;
		/*
		 * Operator << for class ObservationAbstract.
		 * It shows different information of the observation.
		 */
		std::ostream& operator <<(std::ostream & s, ObservationAbstract const & obs) {
			s << "OBSERVATION " << obs.id() << ": of " << obs.typeName() << " from landmark " << obs.landmarkPtr()->typeName() << " and sensor " << obs.sensorPtr()->typeName()
			    << endl;
			s << "Sensor: " << obs.sensorPtr()->id() << ", landmark: " << obs.landmarkPtr()->id() << endl;
			s << " .expectation:  " << obs.expectation << endl;
			s << " .measurement:  " << obs.measurement << endl;
			s << " .innovation:   " << obs.innovation << endl;
			s << " .ev: | prj: " << obs.events.predicted
					<< " | vis: " << obs.events.visible
					<< " | mea: " << obs.events.measured
					<< " | mch: " << obs.events.matched
					<< " | upd: " << obs.events.updated << " | " << endl;
			s << " .cnt:| search: " << obs.counters.nSearch
					<< " | match: " << obs.counters.nMatch
					<< " | inlier: " << obs.counters.nInlier << " | ";
			s << endl << " .visibility: " << obs.landmark().visibilityMap; // FIXME with different sensors
			return s;
		}
		
		image::oimstream& operator <<(image::oimstream & s, ObservationAbstract const & obs) {
			obs.desc_image(s);
			return s;
		}

		ObservationAbstract::ObservationAbstract(const sensor_ptr_t & _senPtr, const landmark_ptr_t & _lmkPtr,
		    const size_t _size_meas, const size_t _size_exp, const size_t _size_inn, const size_t _size_nonobs) :
		    expectation(_size_exp, _size_nonobs),
		    measurement(_size_meas),
		    innovation(_size_inn),
		    prior(_size_nonobs),
		    ia_rsl(ublasExtra::ia_union(_senPtr->ia_globalPose, _lmkPtr->state.ia())),
		    EXP_sg(_size_exp, 7),
		    EXP_l(_size_exp, _lmkPtr->state.size()),
		    EXP_rsl(_size_exp, ia_rsl.size()),
		    INN_meas(_size_inn, _size_meas),
		    INN_exp(_size_inn, _size_exp),
		    INN_rsl(_size_inn, ia_rsl.size()),
		    LMK_sg(_lmkPtr->state.size(),7),
		    LMK_meas(_lmkPtr->state.size(),_size_meas),
		    LMK_prior(_lmkPtr->state.size(),_size_nonobs),
		    LMK_rs(_lmkPtr->state.size(),_senPtr->ia_globalPose.size())
		{
			clearCounters();
			clearFlags();
			updatable = true;
			searchSize = 0;
			clear_hypotheses = true;
		}

		ObservationAbstract::ObservationAbstract(const sensor_ptr_t & _senPtr, const landmark_ptr_t & _lmkPtr,
		    const size_t _size, const size_t _size_nonobs) :
		    expectation(_size, _size_nonobs),
		    measurement(_size),
		    innovation(_size),
		    prior(_size_nonobs),
		    noiseCovariance(_size),
		    ia_rsl(ublasExtra::ia_union(_senPtr->ia_globalPose, _lmkPtr->state.ia())),
		    SG_rs(7, _senPtr->ia_globalPose.size()),
		    EXP_sg(_size, 7),
		    EXP_l(_size, _lmkPtr->state.size()),
		    EXP_rsl(_size, ia_rsl.size()),
		    INN_meas(_size, _size),
		    INN_exp(_size, _size),
		    INN_rsl(_size, ia_rsl.size()),
		    LMK_sg(_lmkPtr->state.size(),7),
		    LMK_meas(_lmkPtr->state.size(),_size),
		    LMK_prior(_lmkPtr->state.size(),_size_nonobs),
		    LMK_rs(_lmkPtr->state.size(),_senPtr->ia_globalPose.size())
		{
	    category = OBSERVATION;
			id(_lmkPtr->id());
			clearCounters();
			clearFlags();
			searchSize = 0;
			clear_hypotheses = true;
		}

		ObservationAbstract::~ObservationAbstract() {
//			cout << "Deleted observation: " << id() << ": " << typeName() << endl;
		}


////		void ObservationAbstract::setup(const feature_ptr_t & featPtr, const Gaussian & _prior){
//		void ObservationAbstract::setup(double _noiseStd, const Gaussian & _prior){
//			noiseCovariance.clear();
//			_noiseStd *= _noiseStd;
//			for (size_t i = 0; i < measurement.size(); i++){
//				noiseCovariance(i,i) = _noiseStd;
//			}
//			measurement.clear();
//			measurement.P(noiseCovariance);
//			prior.x(_prior.x());
//			prior.P(_prior.P());
//		}

		//		void ObservationAbstract::setup(const feature_ptr_t & featPtr, const Gaussian & _prior){
				void ObservationAbstract::setPrior(const Gaussian & _prior) {
					prior = _prior;
//					prior.x(_prior.x());
//					prior.P(_prior.P());
				}

		bool ObservationAbstract::findBounds(jblas::vec & lmk, unsigned iid, jblas::vec7 const & sg, jblas::vec & id_bounds)
		{
			unsigned esize = expectation.size();
			jblas::vec exp(esize), expa(esize), expb(esize), expc(esize), nobs(prior.size());
			double id = lmk(iid);
			model->project_func(sg, lmk, exp, nobs);
			bool changed = false;
/*
			// left extremity
			lmk(iid) = id_bounds(0);
			model->project_func(sg, lmk, expa, nobs);
			if (!model->predictVisibility_func(expa, nobs))
			{ // dichotomy
				double a = id_bounds(0), b = id, c; expb = exp;
				while (ublas::norm_2(expa-expb) > 5)
				{
					c = (a+b)/2;
					model->project_func(sg, lmk, expc, nobs);
					if (model->predictVisibility_func(expc, nobs))
						{ b = c; expb = expc; } else
						{ a = c; expa = expc; }
				}
				id_bounds(0) = b;
				changed = true;
			}

			// right extremity
			lmk(iid) = id_bounds(1);
			model->project_func(sg, lmk, expb, nobs);
			if (!model->predictVisibility_func(expb, nobs))
			{ // dichotomy
				double a = id, b = id_bounds(1), c; expa = exp;
				while (ublas::norm_2(expa-expb) > 5)
				{
					c = (a+b)/2;
					model->project_func(sg, lmk, expc, nobs);
					if (model->predictVisibility_func(expc, nobs))
						{ a = c; expa = expc; } else
						{ b = c; expb = expc; }
				}
				id_bounds(1) = a;
				changed = true;
			}
*/
			for(int i = 0; i < 2; ++i) // for each extremity
			{
				lmk(iid) = id_bounds(i);
				jblas::vec & myexp = (i ? expb : expa);
				model->project_func(sg, lmk, myexp, nobs);
				if (!model->predictVisibility_func(myexp, nobs))
				{ // dichotomy
					double a,b,c;
					if (i) { a=id; b=id_bounds(1); expa = exp; }
						else { a=id_bounds(0); b=id; expb = exp; }
					unsigned n = 0; // add this security because sometimes the distortion correction function is wrong on the borders
					while (ublas::norm_2(expa-expb) > 5 && n < 10)
					{
						c = (a+b)/2;
						lmk(iid) = c;
						model->project_func(sg, lmk, expc, nobs);
						if (model->predictVisibility_func(expc, nobs) == (i == 0))
							{ b = c; expb = expc; } else
							{ a = c; expa = expc; }
						++n;
					}
					if (i) id_bounds(1) = a; else id_bounds(0) = b;
					changed = true;
				}
			}

			lmk(iid) = id;
			return changed;
		}

		bool ObservationAbstract::checkExtremities(jblas::vec & lmk, unsigned iid, double id_sig, Expectation const & hypexp, jblas::vec & id_bounds, jblas::vec & ids, jblas::vec7 const & sg, jblas::vec & exp, jblas::sym_mat & expP, jblas::vec & nobs, double k)
		{
			const double n = 3.0*3.0;

			double id = lmk(iid);
			ids(0) = std::max(id_bounds(0), lmk(iid) - k*id_sig);
			ids(1) = std::min(id_bounds(1), lmk(iid) + k*id_sig);
			bool non_linear = false;
			for(int i = 0; i < 2; ++i)
			{
				lmk(iid) = ids(i);
				model->project_func(sg, lmk, exp, nobs);
				innovation.x() = exp - hypexp.x();
				innovation.P() = expP + hypexp.P();
				if (innovation.mahalanobis() > n) non_linear = true;
			}

			lmk(iid) = id;
			return non_linear;
		}

		void ObservationAbstract::projectHypothesis(Hypothesis & hyp, jblas::vec & lmk, unsigned iid, jblas::vec7 const & sg, jblas::vec & exp, jblas::vec & nobs)
		{
			lmk(iid) = hyp.nobs.x(0);
			ublas::row(hyp_rslP, sensorPtr()->ia_globalPose.size()+iid) = jblas::zero_vec(hyp_rslP.size2());
			hyp_rslP(sensorPtr()->ia_globalPose.size()+iid,sensorPtr()->ia_globalPose.size()+iid) = hyp.nobs.P(0,0); // depth is independent from robot and sensor so no crossed element
			model->project_func(sg, lmk, exp, nobs, hyp.EXP_sg, hyp.EXP_l);
			hyp.expectation.x() = exp;
			mat EXP_rs = prod(hyp.EXP_sg, SG_rs);
			subrange(hyp_EXP_rsl, 0, expectation.size(), 0, sensorPtr()->ia_globalPose.size()) = EXP_rs;
			subrange(hyp_EXP_rsl, 0, expectation.size(), sensorPtr()->ia_globalPose.size(), sensorPtr()->ia_globalPose.size()+landmarkPtr()->state.size()) = hyp.EXP_l;
			hyp.expectation.P() = ublasExtra::prod_JPJt(hyp_rslP, hyp_EXP_rsl);
			//for(unsigned i = 0; i < expectation.size(); ++i) JFR_ASSERT(hyp.expectation.P(i,i) >= 0, "Negative variance");
			hyp.expectation.nonObs = nobs;
		}


		void ObservationAbstract::linearizeUncertainty()
		{
			const double k = 2.5;
			int esize = expectation.size();

			// 1. check non converged landmark
			if (landmark().converged) return;
			// 2. check only one non observable parameter, because we don't know how to do for more FIXME
			jblas::ind_array ind_nobs = landmark().nobs();
			if (ind_nobs.size() != 1) return;
			unsigned inobs = ind_nobs(0);
			// 3. if a decomposition is already computed and it has not been updated, we know we'll need at least as many hypotheses
			bool needs_refinement = false;
			int start_hyp = 2;
			if (clear_hypotheses) { hypotheses.clear(); clear_hypotheses = false; }
			if (hypotheses.size() == 0)
			{
				// 4. check if uncertainty is not too small
				bool big = false;
				for (int i = 0; i < esize; ++i)
					if (expectation.P(0,0) > 20*20)
						{ big = true; break; }
				if (!big) return;
				// 5. check asymetry
				const double ratio = 4.0;
				bool asym = false;
				if (esize == 2)
				{
					double delta = sqrt(jmath::sqr(expectation.P(0,0)-expectation.P(1,1)) + 4*jmath::sqr(expectation.P(0,1)));
					double lambda1 = expectation.P(0,0)+expectation.P(1,1);
					double lambda2 = (lambda1 - delta) / 2.;
					lambda1 = (lambda1 + delta) / 2.;
					lambda1 /= lambda2;
					if (lambda1 > ratio || lambda1 < 1.0/ratio) asym = true;
				} else
				{ // TODO test
					// maybe do something specific for each kind of landmark (eg lines are two points)
					namespace lapack = boost::numeric::bindings::lapack;
					jblas::vec lambda;
					jblas::mat_column_major A(ublas::project(expectation.P(), ublas::range(0,esize), ublas::range(0,esize)));
					jblas::up_sym_adapt_column_major s_A(A);
					int ierr = lapack::syev('V', s_A, lambda, lapack::optimal_workspace());
					if (ierr == 0)
					{
						double min_lambda = lambda(0), max_lambda = lambda(0);
						for(int i = 1; i < esize; ++i) {
							if (lambda(i) < min_lambda) min_lambda = lambda(i);
							if (lambda(i) > max_lambda) max_lambda = lambda(i);
						}
						if (max_lambda / min_lambda > ratio) asym = true;
					} else JFR_WARNING("ObservationAbstract::linearizeUncertainty: error in lapack::syev(), ierr=" << ierr);
				}
				if (!asym) return;
			} else
			{
				needs_refinement = true;
				start_hyp = hypotheses.size();
				if (!events.predicted) start_hyp--; // if first project of frame, to allow decreasing if other sensors make it converge
			}
			// 6. check extremities
			vec lmk = landmarkPtr()->state.x(), exp, nobs;
			hyp_rslP = ublas::project(landmarkPtr()->mapManagerPtr()->mapPtr()->filterPtr->P(), ia_rsl, ia_rsl);
			hyp_EXP_rsl = EXP_rsl;
			sym_mat expP(esize); expP.clear(); //for(int i = 0; i < n; ++i) expP(i) = 1.0; // add 1 pix measurement noise ; actually we're not supposed to
			vec7 sg = sensorPtr()->globalPose();
			vec ids(2);
			double id_sig_0 = sqrt(landmark().state.P(inobs, inobs));
			jblas::vec id_bounds(2); id_bounds(0) = 1e-6; id_bounds(1) = lmk(inobs)+k*id_sig_0;
			if (!needs_refinement) needs_refinement = checkExtremities(lmk, inobs, id_sig_0, expectation, id_bounds, ids, sg, exp, expP, nobs, k);
			if (needs_refinement)
			{
				if (findBounds(lmk, inobs, sg, id_bounds)) // if bounds changed
				{
					needs_refinement = checkExtremities(lmk, inobs, id_sig_0, expectation, id_bounds, ids, sg, exp, expP, nobs, k);
					if (!needs_refinement) hypotheses.clear();
					start_hyp = 2;
				}
			}
			double t = (lmk(inobs)-id_bounds(0)) / id_sig_0;
			double s = (id_bounds(1)-lmk(inobs)) / id_sig_0;

			//-- ok let's decompose !
			for(unsigned n = start_hyp; needs_refinement && n <= max_hypotheses; ++n)
			{
				hypotheses.clear();
				needs_refinement = false;
				double id_sig_n = id_sig_0*(t+s) / (2*n*k - (k-t) - (k-s));
				for(unsigned i = 0; i < n; ++i)
				{
					// build hypothesis
					hypotheses.push_back(Hypothesis(lmk.size(), esize, ind_nobs.size()));
					Hypothesis &hyp = hypotheses.back();
					hyp.nobs = Gaussian(1);
					hyp.nobs.x(0) = (t+2*i*k)*id_sig_n;
					hyp.nobs.P(0,0) = id_sig_n*id_sig_n;
					// project hypothesis
					projectHypothesis(hyp, lmk, inobs, sg, exp, nobs);

					// check hypothesis
					if (checkExtremities(lmk, inobs, id_sig_n, hyp.expectation, id_bounds, ids, sg, exp, expP, nobs, k)) needs_refinement = true;
				}
			}
		}

		void ObservationAbstract::project(bool multi_hyp) {
			// Get global sensor pose
			vec7 sg;
			sensorPtr()->globalPose(sg, SG_rs);

			// project lmk
			vec lmk = landmarkPtr()->state.x();
			vec exp, nobs;
			#if FEJ_EKF
			model->project_func(sg, lmk, exp, nobs, EXP_sg, EXP_l, landmarkPtr()->first_state.x());
			#else
			model->project_func(sg, lmk, exp, nobs, EXP_sg, EXP_l);
			#endif

			// chain rule for Jacobians
			mat EXP_rs = prod(EXP_sg, SG_rs);
			subrange(EXP_rsl, 0, expectation.size(), 0, sensorPtr()->ia_globalPose.size()) = EXP_rs;
			subrange(EXP_rsl, 0, expectation.size(), sensorPtr()->ia_globalPose.size(), sensorPtr()->ia_globalPose.size()+landmarkPtr()->state.size()) = EXP_l;

			// Assignments:
			// x+ = f(x, u, n) :
			expectation.x() = exp;
			// P+ = F_x * P * F_x' + F_n * Q * F_n' :
         expectation.P() = ublasExtra::prod_JPJt(ublas::project(landmarkPtr()->mapManagerPtr()->mapPtr()->filterPtr->P(), ia_rsl, ia_rsl), EXP_rsl);
//         JFR_DEBUG("EXP_rsl \n" << EXP_rsl);
//         JFR_DEBUG("ia_rsl \n" << ia_rsl);
//         JFR_DEBUG("proj \n" << ublas::project(landmarkPtr()->mapManagerPtr()->mapPtr()->filterPtr->P(), ia_rsl, ia_rsl));
//         JFR_DEBUG("expectation \n" << expectation.x() << "\n" << expectation.P());
			// non-observable
			expectation.nonObs = nobs;

// JFR_DEBUG("projected obs " << id() << ": expectation " << expectation.x() << " " << expectation.P());

			if (multi_hyp) linearizeUncertainty();

			// Events
			events.predicted = true;
		}
		
		void ObservationAbstract::projectMean() {
			vec7 sg = sensorPtr()->globalPose();

			vec lmk = landmarkPtr()->state.x();
			vec exp, nobs;
			model->project_func(sg, lmk, exp, nobs);

			expectation.x() = exp;
			expectation.nonObs = nobs;
		}

		void ObservationAbstract::backProject(){
			vec7 sg;

			// Get global sensor pose
			sensorPtr()->globalPose(sg, SG_rs);

			// Make some temporal variables to call function
			vec pix = measurement.x();
			vec invDist = prior.x();
			vec lmk(landmarkPtr()->mySize());
			model->backProject_func(sg, pix, invDist, lmk, LMK_sg, LMK_meas, LMK_prior);

			landmarkPtr()->state.x(lmk);

			LMK_rs = ublas::prod(LMK_sg, SG_rs);

			// Initialize in map
			landmarkPtr()->mapManagerPtr()->mapPtr()->filterPtr
			  ->initialize(
				       landmarkPtr()->mapManagerPtr()->mapPtr()->ia_used_states(),
					LMK_rs,
					sensorPtr()->ia_globalPose,
					landmarkPtr()->state.ia(),
					LMK_meas,
					measurement.P(),
					LMK_prior,
					prior.P());
		}

		void ObservationAbstract::computeInnovation() {
			innovation.x() = measurement.x() - expectation.x();
			innovation.P() = measurement.P() + expectation.P();
			INN_rsl = -EXP_rsl;
		}
		
		void ObservationAbstract::computeInnovationMean(vec &inn, const vec &meas, const vec &exp) const
		{
			inn = meas - exp;
		}

		double ObservationAbstract::computeRelevance() {
			// compute relevance score, maybe a ratio of areas would be better, but this is
			// more simple and more general and should work ok
			innovation.relevance = 0.0;
			for (size_t i = 0; i < measurement.size(); ++i)
				innovation.relevance += expectation.P(i,i)/measurement.P(i,i);
			innovation.relevance /= measurement.size();
			return innovation.relevance;
		}

		void ObservationAbstract::predictInfoGain() {
			expectation.infoGain = ublasExtra::det(expectation.P());
		}

		bool ObservationAbstract::compatibilityTest(double mahaDist){
//JFR_DEBUG("obs " << id() << " passing compatibilityTest with " << mahaDist);
			return (innovation.mahalanobis() < mahaDist*mahaDist);
		}

		bool ObservationAbstract::compatibilityTest(Measurement const & meas, Expectation const & exp, double mahaDist)
		{
			Innovation innov(exp.size());
			innov.x() = meas.x() - exp.x();
			innov.P() = meas.P() + exp.P();
			return (innov.mahalanobis() < mahaDist*mahaDist);
		}

		void ObservationAbstract::clearFlags(){
			int size = sizeof(Events)/sizeof(bool);
			for (int i = 0; i < size; ++i)
				((bool*)&events)[i] = false;
			size = sizeof(Tasks)/sizeof(bool);
			for (int i = 0; i < size; ++i)
				((bool*)&tasks)[i] = false;
		}

		void ObservationAbstract::clearCounters(){
			int size = sizeof(Counters)/sizeof(int);
			for (int i = 0; i < size; ++i)
				((int*)&counters)[i] = 0;
		}

		void ObservationAbstract::update(bool correct_P)
		{
			map_ptr_t mapPtr = sensorPtr()->robotPtr()->mapPtr();
			ind_array ia_x = mapPtr->ia_used_states();

			if (hypotheses.size() != 0)
			{
				// find closest hypothesis
				double min_dist = 300.0*300.0;
				unsigned min_i = 0;
				for(unsigned i = 0; i < hypotheses.size(); ++i)
				{
					innovation.x() = measurement.x() - hypotheses[i].expectation.x();
					innovation.P() = measurement.P() + hypotheses[i].expectation.P();
					double dist = innovation.mahalanobis();
					if (dist < min_dist) { min_dist = dist; min_i = i; }
				}
				// assign this hypothesis
				jblas::ind_array ind_nobs = landmark().state.ia().compose(landmark().nobs()); // indices in the filter space of the non obs elements
				int i = 0; for(jblas::ind_array::const_iterator j = ind_nobs.begin(); j != ind_nobs.end(); ++i, ++j)
				{
					for(jblas::ind_array::const_iterator k = ia_x.begin(); k != ia_x.end(); ++k) mapPtr->filterPtr->P(*j,*k) = 0.0;
					mapPtr->filterPtr->x(*j) = hypotheses[min_i].nobs.x(i);
					mapPtr->filterPtr->P(*j,*j) = hypotheses[min_i].nobs.P(i,i);
				}
				expectation = hypotheses[min_i].expectation;
				EXP_l = hypotheses[min_i].EXP_l;
				ublas::subrange(EXP_rsl, 0, expectation.size(), 0, sensorPtr()->ia_globalPose.size()) = ublas::prod(hypotheses[min_i].EXP_sg, SG_rs);
				ublas::subrange(EXP_rsl, 0, expectation.size(), sensorPtr()->ia_globalPose.size(), sensorPtr()->ia_globalPose.size()+landmarkPtr()->state.size()) = EXP_l;
			}

			// do the update
			computeInnovation();
			mapPtr->filterPtr->correct(ia_x,innovation,INN_rsl,ia_rsl, correct_P) ;
		}
#if 0
		bool ObservationAbstract::voteForKillingLandmark(){
			// kill big ellipses
			// FIXME this is ok for 1 sensor, but not for more, because it won't work with policy ALL
			// probably wee need to fix the policy, and chose a different policy according to the criteria
			// (size = ANY, matchRatio = ALL, consistencyRatio = ...)
			if (events.predicted && events.measured && !events.updated)
			{
				if (searchSize > killSizeTh) {
					JFR_DEBUG( "Obs " << id() << " Killed by size (size " << searchSize << ")" );
					return true;
				}
			}

			// kill unstable and inconsistent lmks
			JFR_ASSERT(counters.nMatch <= counters.nSearch, "counters.nMatch " << counters.nMatch << " > counters.nSearch " << counters.nSearch);
			JFR_ASSERT(counters.nInlier <= counters.nMatch, "counters.nInlier " << counters.nInlier << " > counters.nMatch " << counters.nMatch);
			if (counters.nSearch > killSearchTh) {
				double matchRatio = counters.nMatch / (double) counters.nSearch;
				double consistencyRatio = counters.nInlier / (double)counters.nMatch;

				if (matchRatio < killMatchTh || consistencyRatio < killConsistencyTh)	{
					JFR_DEBUG( "Obs " << id() << " Killed by unstability (match " << matchRatio << " ; consistency " << consistencyRatio << ")" );
					return true;
				}
			}
			return false;
		}
#endif
		void ObservationAbstract::transferInfoObs(observation_ptr_t & obs){
			this->id(obs->id());
			this->name(obs->name());

			this->expectation.x(obs->expectation.x());
			this->expectation.P(obs->expectation.P());
			this->expectation.infoGain = obs->expectation.infoGain;
			this->expectation.nonObs = obs->expectation.nonObs;
			this->expectation.visible = obs->expectation.visible;

			this->innovation.x(obs->innovation.x());
			this->innovation.P(obs->innovation.P());
			this->innovation.iP_ = obs->innovation.iP_;
			this->innovation.mahalanobis_ = obs->innovation.mahalanobis_;

			this->measurement.x(obs->measurement.x());
			this->measurement.P(obs->measurement.P());
			this->measurement.matchScore = obs->measurement.matchScore;

			this->counters = obs->counters;
			this->events = obs->events;
			this->updatable = obs->updatable;
			this->searchSize = obs->searchSize;
		}

	} // namespace rtslam
} // namespace jafar
