/**
 * \file robotConstantVelocity.cpp
 * \date 17/05/2013
 * \author croussil
 * \ingroup rtslam
 */

#include "jmath/jblas.hpp"
#include "jmath/ublasExtra.hpp"
#include "boost/numeric/ublas/operation.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "boost/numeric/ublas/vector_proxy.hpp"
#include "rtslam/quatTools.hpp"

#include "rtslam/robotAbstract.hpp"
#include "rtslam/robotConstantMotionModel.hpp"
#include "rtslam/sensorAbstract.hpp"

namespace jafar {
	namespace rtslam {
		using namespace std;
		using namespace jmath;

		/**
		 * Remote constructor from remote map.
		 * \param _map the remote map
		 */
		template<unsigned LIN_ORDER, unsigned ANG_ORDER>
		RobotConstantMotionModel<LIN_ORDER,ANG_ORDER>::RobotConstantMotionModel(const map_ptr_t & _mapPtr):
			RobotAbstract(_mapPtr)
		{
			// CAUTION: must be consistent with size()
			iaPos = getQuantity(qPos);
			iaOriQuat = getQuantity(qOriQuat);
			if (LIN_ORDER >= 1) {
				registerRobotQuantity(qAbsVel);
				iaAbsVel = getQuantity(qAbsVel); }
			if (ANG_ORDER >= 1) {
				registerRobotQuantity(qAngVel);
				iaAngVel = getQuantity(qAngVel); }
			if (LIN_ORDER >= 2) {
				registerRobotQuantity(qAbsAcc);
				iaAbsAcc = getQuantity(qAbsAcc); }
			if (ANG_ORDER >= 2) {
				registerRobotQuantity(qAngAcc);
				iaAngAcc = getQuantity(qAngAcc); }
			if (LIN_ORDER >= 3) {
				registerRobotQuantity(qAbsJerk);
				iaAbsJerk = getQuantity(qAbsJerk); }
			if (ANG_ORDER >= 3) {
				registerRobotQuantity(qAngJerk);
				iaAngJerk = getQuantity(qAngJerk); }

			JFR_ASSERT(RobotConstantMotionModel::size() == state_size, "wrong state size"); // temporary
			RobotAbstract::construct(_mapPtr, state_size,
										RobotConstantMotionModel::size_control(),
										RobotConstantMotionModel::size_perturbation());

			// Build constant perturbation Jacobian
			constantPerturbation = (ANG_ORDER>0);
			if (constantPerturbation) computePertJacobian();
			#if PERT_DETERMINISTIC
			perturbation.set_deterministic(ublas::range(0,6));
			#endif
			type = CONSTANT_MOTION_MODEL;
		}

		template<unsigned LIN_ORDER, unsigned ANG_ORDER>
		void RobotConstantMotionModel<LIN_ORDER,ANG_ORDER>::move_func(const vec & _x, const vec & _u,
		    const vec & _n, const double _dt, vec & _xnew, mat & _XNEW_x,
				mat & _XNEW_pert, unsigned tempSet) const
		{
			TempVariables & t = (tempSet == 0 ? tempvars0 : tempvars1);
			if (tempSet > 1) std::cerr << "Error: RobotConstantVelocity has only 2 sets of temps variables, increase it if you need more" << std::endl;

			using namespace jblas;
			using namespace ublas;

			/*
			 * This motion model is defined by:
			 * The state vector, x = [p q v w] = [x y z, qw qx qy qz, vx vy vz, wx wy wz], of size 13.
			 * The transition equation x+ = move(x,i), with i = [vi wi] the control impulse, decomposed as:
			 * - p = p + v*dt
			 * - q = q**(w*dt)    <-- ** : quaternion product
			 * - v = v + vi       <-- vi : impulse in linear velocity  - vi = [vix viy viz]
			 * - w = w + wi       <-- wi : impulse in angular velocity - wi = [wix wiy wiz]
			 * -----------------------------------------------------------------------------
			 *
			 * The Jacobian XNEW_x is built with
			 * 						 p     q      v      w      |
			 *						 0     3      7      10     |
			 *      	--------------------------------+------
			 * XNEW_x = [ I_3         PNEW_v        ] | 0  p
			 *       		[      QNEW_q        QNEW_w ] | 3  q
			 *       		[              I_3          ] | 7  v
			 *       		[                      I_3  ] | 10 w
			 * -----------------------------------------------------------------------------
			 *
			 * The Jacobian XNEW_pert is built with
			 *          			 vi     wi   |
			 *                 0      3    |
			 *       			-----------------+------
			 * XNEW_pert = 	[            ] | 0  p
			 * 					   	[            ] | 3  q
			 * 							[ I_3        ] | 7  v
			 * 							[        I_3 ] | 10 w
			 * this Jacobian is however constant and is computed once at Construction time.
			 *
			 * NOTE: The also constant perturbation matrix:
			 *    Q = XNEW_pert * perturbation.P * trans(XNEW_pert)
			 * could be built also once after construction with computeStatePerturbation().
			 * This is up to the user -- if nothing is done, Q will be computed at each iteration.
			 * -----------------------------------------------------------------------------
			 */

			// split robot state vector
			vec3 p, v, w, va, wa, vj, wj;
			vec4 q;
			splitState(_x, p, q, v, w, va, wa, vj, wj);

			// split perturbation vector
			vec3 vi, wi;
			splitControl(_n, vi, wi);

			// Non-trivial Jacobian blocks
			identity_mat I_3(3);

			vec3 pnew, vnew, wnew, vanew, wanew, vjnew, wjnew;
			vec4 qnew, qwdt;


			// predict each part of the state, give or build non-trivial Jacobians
			switch (LIN_ORDER)
			{
				case 3: vanew = va + vj * _dt;
				case 2: vnew  = v  + va * _dt;
				case 1: pnew  = p  + v  * _dt;
				case 0: ;
			}
			switch (ANG_ORDER)
			{
				case 3: wanew = wa + wj * _dt;
				case 2: wnew  = w  + wa * _dt;
				case 1:
					quaternion::v2q(w * _dt, qwdt, t.QWDT_wdt);
					quaternion::qProd(q, qwdt, qnew, t.QNEW_q, t.QNEW_qwdt);
				case 0: ;
			}

			switch (LIN_ORDER)
			{
				case 3: vjnew = vj + vi; break;
				case 2: vanew = va + vi; break;
				case 1: vnew  = v  + vi; break;
				case 0: pnew  = p  + vi; break;
			}		
			switch (ANG_ORDER)
			{
				case 3: wjnew = wj + wi; break;
				case 2: wanew = wa + wi; break;
				case 1: wnew  = w  + wi; break;
				case 0:
					quaternion::v2q(wi, qwdt, t.QWDT_wdt);
					quaternion::qProd(q, qwdt, qnew, t.QNEW_q, t.QNEW_qwdt);
					break;
			}


			// normalize quaternion
			ublasExtra::normalizeJac(qnew, t.QNORM_qnew);
			ublasExtra::normalize(qnew);

			// Compose state - this is the output state.
			unsplitState(pnew, qnew, vnew, wnew, vanew, wanew, vjnew, wjnew, _xnew);

			// Build transition Jacobian matrix XNEW_x
			_XNEW_x.assign(identity_mat(state.size()));
			switch (LIN_ORDER)
			{
				case 3: project(_XNEW_x, iaAbsAcc, iaAbsJerk) = I_3 * _dt;
				case 2: project(_XNEW_x, iaAbsVel, iaAbsAcc) = I_3 * _dt;
				case 1: project(_XNEW_x, iaPos, iaAbsVel) = I_3 * _dt;
				case 0: ;
			}
			switch (ANG_ORDER)
			{
				case 3: project(_XNEW_x, iaAngAcc, iaAngJerk) = I_3 * _dt;
				case 2: project(_XNEW_x, iaAngVel, iaAngAcc) = I_3 * _dt;
				case 1:
					t.QNEW_wdt = prod(t.QNEW_qwdt, t.QWDT_wdt);
					project(_XNEW_x, iaOriQuat, iaOriQuat) = prod(t.QNORM_qnew,t.QNEW_q);
					project(_XNEW_x, iaOriQuat, iaAngVel) = prod(t.QNORM_qnew,t.QNEW_wdt) * _dt;
				case 0: ;
			}


			/*
			 * We are normally supposed here to build the perturbation Jacobian matrix XNEW_pert.
			 * NOTE: XNEW_pert is constant and it has been build in the constructor.
			 *
			 * \sa See computePertJacobian() for more info.
			 */
			if (ANG_ORDER == 0)
			{
				_XNEW_pert.clear();
				t.QNEW_wdt = prod(t.QNEW_qwdt, t.QWDT_wdt);
				project(_XNEW_pert, iaPos, range(0,3)) = I_3;
				project(_XNEW_pert, iaOriQuat, range(3,6)) = prod(t.QNORM_qnew,t.QNEW_wdt);
			}

		}

		/*
		 * Build perturbation Jacobian matrix XNEW_pert.
		 *
		 * The perturbation Jacobian is
		 *
		 * var    |  vi  wi
		 *    pos |  0   3
		 * -------+--------
		 *  p  0  |  0   0
		 *  q  3  |  0   0
		 *  v  7  |  I   0
		 *  w  10 |  0   I
		 *
		 * NOTE: These lines below just for reference:
		 */
		template<unsigned LIN_ORDER, unsigned ANG_ORDER>
		void RobotConstantMotionModel<LIN_ORDER,ANG_ORDER>::computePertJacobian() {
			identity_mat I(3);
			XNEW_pert.clear();

			switch (LIN_ORDER)
			{
				case 3: project(XNEW_pert, iaAbsJerk, range(0, 3)) = I; break;
				case 2: project(XNEW_pert, iaAbsAcc , range(0, 3)) = I; break;
				case 1: project(XNEW_pert, iaAbsVel , range(0, 3)) = I; break;
				case 0: project(XNEW_pert, iaPos    , range(0, 3)) = I; break;
			}
			switch (ANG_ORDER)
			{
				case 3: project(XNEW_pert, iaAngJerk, range(3, 6)) = I; break;
				case 2: project(XNEW_pert, iaAngAcc , range(3, 6)) = I; break;
				case 1: project(XNEW_pert, iaAngVel , range(3, 6)) = I; break;
				case 0: ; // not constant so computed in move_func
			}
		}


		template<unsigned LIN_ORDER, unsigned ANG_ORDER>
		void RobotConstantMotionModel<LIN_ORDER,ANG_ORDER>::writeLogHeader(kernel::DataLogger& log) const
		{
			std::ostringstream oss; oss << "Robot " << id();
			log.writeComment(oss.str());
			
			log.writeLegendTokens("time");
			log.writeLegendTokens("sensor");

			log.writeLegendTokens("absx absy absz");
			log.writeLegendTokens("absyaw abspitch absroll");
			log.writeLegendTokens("x y z");
			log.writeLegendTokens("qw qx qy qz");
			if (LIN_ORDER>=1) log.writeLegendTokens("vx vy vz");
			if (ANG_ORDER>=1) log.writeLegendTokens("vyaw vpitch vroll");
			if (LIN_ORDER>=2) log.writeLegendTokens("ax ay az");
			if (ANG_ORDER>=2) log.writeLegendTokens("ayaw apitch aroll");
			if (LIN_ORDER>=3) log.writeLegendTokens("jx jy jz");
			if (ANG_ORDER>=3) log.writeLegendTokens("jyaw jpitch jroll");

			log.writeLegendTokens("sig_absx sig_absy sig_absz");
			log.writeLegendTokens("sig_absyaw sig_abspitch sig_absroll");
			log.writeLegendTokens("sig_x sig_y sig_z");
			log.writeLegendTokens("sig_qw sig_qx sig_qy sig_qz");
			if (LIN_ORDER>=1) log.writeLegendTokens("sig_vx sig_vy sig_vz");
			if (ANG_ORDER>=1) log.writeLegendTokens("sig_vyaw sig_vpitch sig_vroll");
			if (LIN_ORDER>=2) log.writeLegendTokens("sig_ax sig_ay sig_az");
			if (ANG_ORDER>=2) log.writeLegendTokens("sig_ayaw sig_apitch sig_aroll");
			if (LIN_ORDER>=3) log.writeLegendTokens("sig_ax sig_ay sig_az");
			if (ANG_ORDER>=3) log.writeLegendTokens("sig_ayaw sig_apitch sig_aroll");

			log.writeLegendTokens("cov_xy cov_xz cov_yz");
			log.writeLegendTokens("cov_qwqx cov_qwqy cov_qwqz cov_qxqy cov_qxqz cov_qyqz");
			if (LIN_ORDER>=1) log.writeLegendTokens("cov_vxvy cov_vxvz cov_vyvz");
		}
		
		template<unsigned LIN_ORDER, unsigned ANG_ORDER>
		void RobotConstantMotionModel<LIN_ORDER,ANG_ORDER>::writeLogData(kernel::DataLogger& log) const
		{
			jblas::vec state_x(6); jblas::sym_mat state_P(6,6);
			jblas::vec stateq_x = pose.x();
			jblas::sym_mat stateq_P = pose.P();
			slamPoseToRobotPose(stateq_x, stateq_P, state_x, state_P);

			log.writeData(self_time);
			log.writeData(last_updated->name());

			for(unsigned i = 0 ; i < 3 ; ++i) log.writeData(state_x(i));
			for(unsigned i = 0 ; i < 3 ; ++i) log.writeData(state_x(3+2-i));
			for(unsigned i = 0 ; i < 7; ++i) log.writeData(stateq_x(i));
			if (LIN_ORDER>=1) for(unsigned i = iaAbsVel.start(); i < iaAbsVel.start()+iaAbsVel.size(); ++i) log.writeData(state.x()(i));
			if (ANG_ORDER>=1) for(unsigned i = iaAngVel.start(); i < iaAngVel.start()+iaAngVel.size(); ++i) log.writeData(state.x()((int)(2-i)+2*iaAngVel.start()));
			if (LIN_ORDER>=2) for(unsigned i = iaAbsAcc.start(); i < iaAbsAcc.start()+iaAbsAcc.size(); ++i) log.writeData(state.x()(i));
			if (ANG_ORDER>=2) for(unsigned i = iaAngAcc.start(); i < iaAngAcc.start()+iaAngAcc.size(); ++i) log.writeData(state.x()((int)(2-i)+2*iaAngAcc.start()));
			if (LIN_ORDER>=3) for(unsigned i = iaAbsJerk.start(); i < iaAbsJerk.start()+iaAbsJerk.size(); ++i) log.writeData(state.x()(i));
			if (ANG_ORDER>=3) for(unsigned i = iaAngJerk.start(); i < iaAngJerk.start()+iaAngJerk.size(); ++i) log.writeData(state.x()((int)(2-i)+2*iaAngJerk.start()));

			for(unsigned i = 0 ; i < 3 ; ++i) log.writeData(state_P(i,i));
			for(unsigned i = 0 ; i < 3 ; ++i) log.writeData(state_P(3+2-i,3+2-i));
			for(unsigned i = 0; i < 7; ++i) log.writeData(sqrt(stateq_P(i,i)));
			if (LIN_ORDER>=1) for(unsigned i = iaAbsVel.start(); i < iaAbsVel.start()+iaAbsVel.size(); ++i) log.writeData(sqrt(state.P()(i,i)));
			if (ANG_ORDER>=1) for(unsigned i = iaAngVel.start(); i < iaAngVel.start()+iaAngVel.size(); ++i) { int idx = (int)(2-i)+2*iaAngVel.start(); log.writeData(sqrt(state.P()(idx,idx))); }
			if (LIN_ORDER>=2) for(unsigned i = iaAbsAcc.start(); i < iaAbsAcc.start()+iaAbsAcc.size(); ++i) log.writeData(sqrt(state.P()(i,i)));
			if (ANG_ORDER>=2) for(unsigned i = iaAngAcc.start(); i < iaAngAcc.start()+iaAngAcc.size(); ++i) { int idx = (int)(2-i)+2*iaAngAcc.start(); log.writeData(sqrt(state.P()(idx,idx))); }
			if (LIN_ORDER>=3) for(unsigned i = iaAbsJerk.start(); i < iaAbsJerk.start()+iaAbsJerk.size(); ++i) log.writeData(sqrt(state.P()(i,i)));
			if (ANG_ORDER>=3) for(unsigned i = iaAngJerk.start(); i < iaAngJerk.start()+iaAngJerk.size(); ++i) { int idx = (int)(2-i)+2*iaAngJerk.start(); log.writeData(sqrt(state.P()(idx,idx))); }

			for(unsigned i = 0; i < 3; ++i) for(unsigned j = i+1; j < 3; ++j) log.writeData(stateq_P(i,j));
			for(unsigned i = 0; i < 4; ++i) for(unsigned j = i+1; j < 4; ++j) log.writeData(stateq_P(3+i,3+j));
			if (LIN_ORDER>=1) for(unsigned i = 0; i < 3; ++i) for(unsigned j = i+1; j < 3; ++j) log.writeData(state.P()(iaAbsVel.start()+i,iaAbsVel.start()+j));
		}


		template class RobotConstantMotionModel<0,0>;
		template class RobotConstantMotionModel<1,1>;
		template class RobotConstantMotionModel<2,2>;
		template class RobotConstantMotionModel<3,3>;

		template class RobotConstantMotionModel<2,1>;

	}
}
