/**
 * \file sensorAbstract.cpp
 * \date 10/03/2010
 * \author jsola
 * \ingroup rtslam
 */

#include "boost/assign/std/vector.hpp"
#include "boost/shared_ptr.hpp"
#include "jmath/indirectArray.hpp"
#include "rtslam/sensorAbstract.hpp"
#include "rtslam/robotAbstract.hpp"
#include "rtslam/observationAbstract.hpp"
#include "rtslam/quatTools.hpp"

#include "jmath/angle.hpp"
#include <vector>


namespace jafar {
	namespace rtslam {
		using namespace std;
		using namespace boost;
		using namespace ublas;
		using namespace ublasExtra;

		IdFactory SensorAbstract::sensorIds = IdFactory();

		/*
		 * Operator << for class SensorAbstract.
		 * It shows different information of the sensor.
		 */
		std::ostream& operator <<(std::ostream & s,
		    SensorAbstract const & sen) {
			s << sen.categoryName() << " " << sen.id() << ": ";
			if (sen.name().size() > 0) s << sen.name() << ", ";
			s << "of type " << sen.typeName() << std::endl;
			s << ".pose :  " << sen.pose << endl;
			if (sen.state.storage() == Gaussian::REMOTE) s << endl << ".ia_s: "
			    << sen.state.ia();
			s << ".in robot: " << sen.robot().id();
			return s;
		}

		SensorAbstract::SensorAbstract(const robot_ptr_t & _robPtr, const filtered_obj_t poseInFilter, int extraStateFilterSize,
			std::vector<RobotAbstract::Quantity> robQuant):
			MapObject(_robPtr->mapPtr(), (poseInFilter == FILTERED ? 7 : 0) + extraStateFilterSize, (poseInFilter == FILTERED ? 0 : 7)),
			integrate_all(false), use_for_init(false), need_init(false), enabled(true),
			quantsJacob(robQuant), quantsState(*_robPtr, robQuant),
			pose((poseInFilter == FILTERED ? state : stateU), Gaussian::REMOTE),
			ia_globalPose(ia_union(_robPtr->state.ia().compose(quantsState.getQuantities()), state.ia())),
			missed_data(0)
		{
			category = SENSOR;
			this->poseInFilter = (poseInFilter == FILTERED);
			setId();
		}

		void SensorAbstract::setPose(double x, double y, double z, double rollDeg,
		    double pitchDeg, double yawDeg) {
			const double pos_[3] = { x, y, z };
			const double euler_[3] = { jmath::degToRad(rollDeg),
			    jmath::degToRad(pitchDeg), jmath::degToRad(yawDeg) };
			ublas::subrange(pose.x(), 0, 3) = createVector<3> (pos_);
			ublas::subrange(pose.x(), 3, 7)
			    = quaternion::e2q(createVector<3> (euler_));
		}

		void SensorAbstract::setPoseStd(double x, double y, double z, double rollDeg,
		    double pitchDeg, double yawDeg, double xStd, double yStd, double zStd,
		    double rollDegStd, double pitchDegStd, double yawDegStd) {

			const double pos_[3] = { x, y, z };
			const double euler_[3] = { jmath::degToRad(rollDeg), jmath::degToRad(pitchDeg), jmath::degToRad(yawDeg) };

			// convert euler pose to quat pose
			ublas::subrange(pose.x(), 0, 3) = createVector<3> (pos_);

			vec3 euler = createVector<3> (euler_);
			ublas::subrange(pose.x(), 3, 7) = quaternion::e2q(euler);

			// convert euler uncertainty to quaternion uncertainty
			const double posStd_[3] = { xStd, yStd, zStd };
			const double eulerStd_[3] = { jmath::degToRad(rollDegStd),
			    jmath::degToRad(pitchDegStd), jmath::degToRad(yawDegStd) };

			vec3 eulerStd = createVector<3> (eulerStd_);

			Gaussian E(3);	E.std(eulerStd);
			vec4 q;
			mat Q_e(4, 3);

			quaternion::e2q(euler, q, Q_e);

			// write pose
			subrange(pose.P(), 0,3, 0,3) = createSymMat<3>(posStd_);
			subrange(pose.P(), 3,7, 3,7) = prod(Q_e, prod<mat>(E.P(), trans(Q_e)));

		}


		/*
		 * Get sensor pose in global frame.
		 */
		vec7 SensorAbstract::globalPose(bool normalize) {
			vec7 globPose;
			jblas::vec7 robotPose = robotPtr()->pose.x();
			jblas::vec7 sensorPose = pose.x();
			if (normalize)
			{
				ublas::subrange(robotPose,3,7) = ublasExtra::normalized(ublas::subrange(robotPose,3,7));
				if (poseInFilter) ublas::subrange(sensorPose,3,7) = ublasExtra::normalized(ublas::subrange(sensorPose,3,7));
			}
			globPose = quaternion::composeFrames(robotPose, sensorPose);
			return globPose;
		}

		/*
		 * Get sensor pose in global frame.
		 */
		void SensorAbstract::globalPose(jblas::vec7 & senGlobalPos,
		    jblas::mat & SG_rs) {
			using boost::numeric::ublas::range;
			jblas::vec7 robotPose = robotPtr()->pose.x();
			jblas::vec7 sensorPose = pose.x();

			if (!poseInFilter) {
				// Sensor is not in the map. Jacobian only wrt robot.
				senGlobalPos = quaternion::composeFrames(robotPose, sensorPose);
				quaternion::composeFrames_by_dglobal(robotPose, sensorPose, SG_rs);
			} else {
				// Sensor is in the map. Give composed Jacobian.
				jblas::mat PG_r(7, 7), PG_s(7, 7);
				quaternion::composeFrames(robotPose, sensorPose, senGlobalPos, PG_r,
				                          PG_s);
				project(SG_rs, range(0, 7), range(0, 7)) = PG_r;
				project(SG_rs, range(0, 7), range(7, 14)) = PG_s;
			}
		}
		
		void SensorExteroAbstract::process(unsigned id, double date_limit)
		{
			// get data
			hardwareSensorPtr->getRaw(id, rawPtr);
			rawCounter++;
			rawPtr->id() = rawCounter;
			
			// observe
			for (DataManagerList::iterator dmaIter = dataManagerList().begin(); dmaIter != dataManagerList().end(); ++dmaIter)
			{
				// TODO enable time management in detectNew as well, and estimate how much time will be needed by manage and detectNew
				const double time_manage_detect = 0.000; // try nothing for now because processKnown will often stop before
				data_manager_ptr_t dmaPtr = *dmaIter;
				double date_limit_process = date_limit - time_manage_detect;
				double date_now = kernel::Clock::getTime();
				bool zap_process = false;
				if (date_limit < 0 || date_now < date_limit_process)
					dmaPtr->processKnown(rawPtr, date_limit_process);  else zap_process = true;
				date_now = kernel::Clock::getTime();
				if (!zap_process || date_now < date_limit)
					dmaPtr->mapManagerPtr()->manage();
				date_now = kernel::Clock::getTime();
				if (!zap_process || date_now < date_limit)
					dmaPtr->detectNew(rawPtr);
				dmaPtr->dump(rawPtr);
			}
			
			//hardwareSensorPtr->release();
		}



		void SensorAbstract::writeLogHeader(kernel::DataLogger& log) const
		{
			std::ostringstream oss; oss << "Sensor " << id();
			log.writeComment(oss.str());
			if (poseInFilter)
			{
				log.writeLegendTokens("x y z");
				log.writeLegendTokens("qw qx qy qz");
				log.writeLegendTokens("yaw pitch roll");
				log.writeLegendTokens("sig_x sig_y sig_z");
				log.writeLegendTokens("sig_qw sig_qx sig_qy sig_qz");
				log.writeLegendTokens("sig_yaw sig_pitch sig_roll");
			}
		}

		void SensorAbstract::writeLogData(kernel::DataLogger& log) const
		{
			if (poseInFilter)
			{
				jblas::vec4 q = ublas::subrange(pose.x(),3,7);
				jblas::mat Q_q(4,4); jblas::sym_mat Q(4,4);
				ublasExtra::normalizeJac(q, Q_q);
				ublasExtra::normalize(q);
				Q = ublasExtra::prod_JPJt(ublas::subrange(pose.P(), 3,7, 3,7), Q_q);

				jblas::vec euler(3); jblas::sym_mat eulerP(3,3);
				quaternion::q2e(q, Q, euler, eulerP);
				for(unsigned i = 0; i < 3; ++i) log.writeData(pose.x()(i));
				for(unsigned i = 0; i < 4; ++i) log.writeData(q(i));
				for(unsigned i = 0; i < 3 ; ++i) log.writeData(euler(2-i));
				for(unsigned i = 0; i < 3; ++i) log.writeData(sqrt(pose.P()(i,i)));
				for(unsigned i = 0; i < 4; ++i) log.writeData(sqrt(Q(i,i)));
				for(unsigned i = 0; i < 3 ; ++i) log.writeData(sqrt(eulerP(2-i,2-i)));
			}
		}

	}
}
